#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os
from datetime import datetime

from commoncode.system import on_posix
from commoncode.functional import memoize

"""
Low level file type utilities, essentially a wrapper around os.path and stat.
"""


def is_link(location):
    """
    Return True if `location` is a symbolic link.
    """
    return location and os.path.islink(location)


def is_file(location, follow_symlinks=False):
    """
    Return True if `location` is a file.
    """
    _is_file = location and os.path.isfile(location)
    if follow_symlinks:
        return _is_file
    return _is_file and not is_link(location) and not is_broken_link(location)


def is_dir(location, follow_symlinks=False):
    """
    Return True if `location` is a directory.
    """
    _is_dir = location and os.path.isdir(location) and not is_file(location)
    if follow_symlinks:
        return _is_dir
    return _is_dir and not is_link(location) and not is_broken_link(location)


def is_regular(location):
    """
    Return True if `location` is regular. A regular location is a file or a
    dir and not a special file or symlink.
    """
    return location and (is_file(location) or is_dir(location))


def is_special(location):
    """
    Return True if `location` is a special file . A special file is not a
    regular file, i.e. anything such as a broken link, block file, fifo,
    socket, character device or else.
    """
    return not is_regular(location)


def is_broken_link(location):
    """
    Return True if `location` is a broken link.
    """
    # always false on windows, until Python supports junctions/links
    if on_posix and is_link(location):
        target = get_link_target(location)
        target_loc = os.path.join(os.path.dirname(location), target)
        return target and not os.path.exists(target_loc)


def get_link_target(location):
    """
    Return the link target for `location` if this is a Link or an empty
    string.
    """
    target = ''
    # always false on windows, until Python supports junctions/links
    if on_posix and is_link(location):
        try:
            # return false on OSes not supporting links
            target = os.readlink(location)
        except UnicodeEncodeError:
            # location is unicode but readlink can fail in some cases
            pass
    return target


# Map of type checker function -> short type code
# The order of types check matters: link -> file -> directory -> special
TYPES = dict([
    (is_link, ('l', 'link',)),
    (is_file, ('f', 'file',)),
    (is_dir, ('d', 'directory',)),
    (is_special, ('s', 'special',))
])


def get_type(location, short=True):
    """
    Return the type of the `location` or None if it does not exist.
    Return the short form (single character) or long form if short=False
    """
    if location:
        for type_checker in TYPES:
            tc = type_checker(location)
            if tc:
                short_form, long_form = TYPES[type_checker]
                return short and short_form or long_form


def is_readable(location):
    """
    Return True if the file at location has readable permission set.
    Does not follow links.
    """
    if location:
        if is_dir(location):
            return os.access(location, os.R_OK | os.X_OK)
        else:
            return os.access(location, os.R_OK)


def is_writable(location):
    """
    Return True if the file at location has writeable permission set.
    Does not follow links.
    """
    if location:
        if is_dir(location):
            return os.access(location, os.R_OK | os.W_OK | os.X_OK)
        else:
            return os.access(location, os.R_OK | os.W_OK)


def is_executable(location):
    """
    Return True if the file at location has executable permission set.
    Does not follow links.
    """
    if location:
        if is_dir(location):
            return os.access(location, os.R_OK | os.W_OK | os.X_OK)
        else:
            return os.access(location, os.X_OK)


def is_rwx(location):
    """
    Return True if the file at location has read, write and executable
    permission set. Does not follow links.
    """
    return is_readable(location) and is_writable(location) and is_executable(location)


def get_last_modified_date(location):
    """
    Return the last modified date stamp of a file as YYYYMMDD format. The date
    of non-files (dir, links, special) is always an empty string.
    """
    yyyymmdd = ''
    if is_file(location):
        utc_date = datetime.isoformat(
            datetime.utcfromtimestamp(os.path.getmtime(location))
        )
        yyyymmdd = utc_date[:10]
    return yyyymmdd


counting_functions = {
    'file_count': lambda _: 1,
    'file_size': os.path.getsize,
}


@memoize
def counter(location, counting_function):
    """
    Return a count for a single file or a cumulative count for a directory
    tree at `location`.

    Get a callable from the counting_functions registry using the
    `counting_function` string. Call this callable with a `location` argument
    to determine the count value for a single file. This allow memoization
    with hashable arguments.

    Only regular files and directories have a count. The count for a directory
    is the recursive count sum of the directory file and directory
    descendants.

    Any other file type such as a special file or link has a zero size. Does
    not follow links.
    """
    if not (is_file(location) or is_dir(location)):
        return 0

    count = 0
    if is_file(location):
        count_fun = counting_functions[counting_function]
        return count_fun(location)
    elif is_dir(location):
        count += sum(counter(os.path.join(location, p), counting_function)
                     for p in os.listdir(location))
    return count


def get_file_count(location):
    """
    Return the cumulative number of files in the directory tree at `location`
    or 1 if `location` is a file. Only regular files are counted. Everything
    else has a zero size.
    """
    return counter(location, 'file_count')


def get_size(location):
    """
    Return the size in bytes of a file at `location` or if `location` is a
    directory, the cumulative size of all files in this directory tree. Only
    regular files have a size. Everything else has a zero size.
    """
    return counter(location, 'file_size')



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import fnmatch
import os

from commoncode import fileutils
from commoncode import paths

TRACE = False
if TRACE:
    import logging
    import sys

    logger = logging.getLogger(__name__)
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
    logger.setLevel(logging.DEBUG)

"""
Match files and directories paths based on inclusion and exclusion glob-style
patterns.

For example, this can be used to skip files that match ignore patterns,
similar to a version control ignore files such as .gitignore.

The pattern syntax is the same as fnmatch(5) as implemented in Python.

Patterns are applied to a path this way:
 - Paths are converted to POSIX paths before matching.
 - Patterns are NOT case-sensitive.
 - Leading slashes are ignored.
 - If the pattern contains a /, then the whole path must be matched;
   otherwise, the pattern matches if any path segment matches.
 - When matched, a directory content is matched recursively.
   For instance, when using patterns for ignoring, a matched directory will
   be ignored with its file and sub-directories at full depth.
 - The order of patterns does not matter, except for exclusions vs. inclusions.
 - Exclusion patterns are prefixed with an exclamation mark (bang or !)
   meaning that matched paths by that pattern will be excluded. Exclusions
   have precedence of inclusions.
 - Patterns starting with # are comments and skipped. use [#] for a literal #.
 - to match paths relative to some root path, you must design your patterns
   and the paths to be tested accordingly. This module does not handles this.

Patterns may include glob wildcards such as:
 - ? : matches any single character.
 - * : matches 0 or more characters.
 - [seq] : matches any character in seq
 - [!seq] :matches any character not in seq
For a literal match, wrap the meta-characters in brackets. For example, '[?]'
matches the character '?'.
"""


def is_included(path, includes=None, excludes=None):
    """
    Return a True if `path` is included based on mapping of `includes` and
    `excludes` glob patterns. If the `path` is empty, return False.

    Matching is done based on the set of `includes` and `excludes` patterns maps
    of {fnmatch pattern: message}. If `includes` are provided they are tested
    first. The `excludes` are tested second if provided.

    The ordering of the includes and excludes items does not matter and if a map
    is empty, it is not used for matching.
    """
    if not path or not path.strip():
        return False

    if not includes and not excludes:
        return True

    includes = includes or {}
    includes = {k: v for k, v in includes.items() if k}
    excludes = excludes or {}
    excludes = {k: v for k, v in excludes.items() if k}

    if includes:
        included = get_matches(path, includes, all_matches=False)
        if TRACE:
            logger.debug('in_fileset: path: %(path)r included:%(included)r' % locals())
        if not included:
            return False

    if excludes:
        excluded = get_matches(path, excludes, all_matches=False)
        if TRACE:
            logger.debug('in_fileset: path: %(path)r excluded:%(excluded)r .' % locals())
        if excluded:
            return False

    return True


def get_matches(path, patterns, all_matches=False):
    """
    Return a list of values (which are values from the matched `patterns`
    mappint of {pattern: value or message} if `path` is matched by any of the
    pattern from the `patterns` map or an empty list.
    If `all_matches` is False, stop and return on the first matched pattern.
    """
    if not path or not patterns:
        return False

    path = fileutils.as_posixpath(path).lower()
    pathstripped = path.lstrip('/0')
    if not pathstripped:
        return False

    segments = paths.split(pathstripped)

    if TRACE:
        logger.debug('_match: path: %(path)r patterns:%(patterns)r.' % locals())

    matches = []
    if not isinstance(patterns, dict):
        assert isinstance(patterns, (list, tuple)), 'Invalid patterns: {}'.format(patterns)
        patterns = {p: p for p in patterns}

    for pat, value in patterns.items():
        if not pat or not pat.strip():
            continue

        value = value or ''
        pat = pat.lstrip('/').lower()
        is_plain = '/' not in pat

        if is_plain:
            if any(fnmatch.fnmatchcase(s, pat) for s in segments):
                matches.append(value)
                if not all_matches:
                    break
        elif (fnmatch.fnmatchcase(path, pat) or fnmatch.fnmatchcase(pathstripped, pat)):
            matches.append(value)
            if not all_matches:
                break
    if TRACE:
        logger.debug('_match: matches: %(matches)r' % locals())

    if not all_matches:
        if matches:
            return matches[0]
        else:
            return False
    return matches


def load(location):
    """
    Return a sequence of patterns from a file at location.
    """
    if not location:
        return tuple()
    fn = os.path.abspath(os.path.normpath(os.path.expanduser(location)))
    msg = ('File %(location)s does not exist or not a file.') % locals()
    assert (os.path.exists(fn) and os.path.isfile(fn)), msg
    mode = 'r'
    with open(fn, mode) as f:
        return [l.strip() for l in f if l and l.strip()]


def includes_excludes(patterns, message):
    """
    Return a dict of included patterns and a dict of excluded patterns from a
    sequence of `patterns` strings and a `message` setting the message as
    value in the returned mappings. Ignore pattern as comments if prefixed
    with #. Use an empty string is message is None.
    """
    message = message or ''
    BANG = '!'
    POUND = '#'
    included = {}
    excluded = {}
    if not patterns:
        return included, excluded

    for pat in patterns:
        pat = pat.strip()
        if not pat or pat.startswith(POUND):
            continue
        if pat.startswith(BANG):
            cpat = pat.lstrip(BANG)
            if cpat:
                excluded[cpat] = message
            continue
        else:
            included.add[pat] = message
    return included, excluded



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import json
import os
import sys
import traceback
from collections import deque
from functools import partial
from hashlib import md5
from operator import itemgetter
from os import walk as os_walk
from os.path import abspath
from os.path import exists
from os.path import expanduser
from os.path import isfile
from os.path import join
from os.path import normpath
from posixpath import join as posixpath_join
from posixpath import normpath as posixpath_normpath
from posixpath import dirname as posixpath_parent

import attr

try:
    from scancode_config import scancode_temp_dir as temp_dir
except ImportError:
    # alway have something there.
    import tempfile

    temp_dir = tempfile.mkdtemp(prefix='scancode-resource-cache')

from commoncode import ignore
from commoncode.datautils import List
from commoncode.datautils import Mapping
from commoncode.datautils import String
from commoncode.filetype import is_file as filetype_is_file
from commoncode.filetype import is_special
from commoncode.fileutils import as_posixpath
from commoncode.fileutils import create_dir
from commoncode.fileutils import delete
from commoncode.fileutils import file_name
from commoncode.fileutils import parent_directory
from commoncode.fileutils import splitext_name

"""
This module provides Codebase and Resource objects as an abstraction for files
and directories used throughout ScanCode. ScanCode deals with a lot of these as
they are the basic unit of processing.

A Codebase is a tree of Resource. A Resource represents a file or directory and
holds essential file information as attributes. At runtime, scan data is added
as attributes to a Resource. Resource are kept in memory or saved on disk.

This module handles all the details of walking files, path handling and caching.
"""

# Tracing flags
TRACE = False
TRACE_DEEP = False


def logger_debug(*args):
    pass


if TRACE or TRACE_DEEP:
    import logging

    logger = logging.getLogger(__name__)
    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)

    def logger_debug(*args):
        return logger.debug(' '.join(isinstance(a, str) and a or repr(a) for a in args))


class ResourceNotInCache(Exception):
    pass


class UnknownResource(Exception):
    pass


def skip_ignored(location):
    """
    Return True if ``location`` should be skipped.
    Always ignore VCS and some special filetypes.
    """
    ignored = partial(ignore.is_ignored, ignores=ignore.ignores_VCS)

    if TRACE_DEEP:
        logger_debug()
        logger_debug(
            'Codebase.populate: walk: ignored loc:',
            location,
            'ignored:',
            ignored(location),
            'is_special:',
            is_special(location),
        )

    return is_special(location) or ignored(location)


def depth_walk(
    root_location,
    max_depth,
    skip_ignored=skip_ignored,
    error_handler=lambda: None,
):
    """
    Yield a (top, dirs, files) tuple at each step of walking the ``root_location``
    directory recursively up to ``max_depth`` path segments extending from the
    ``root_location``. The behaviour is similar of ``os.walk``.

    Arguments:

    - root_location: Absolute, normalized path for the directory to be walked
    - max_depth: positive integer for fixed depth limit. 0 for no limit.
    - skip_ignored: Callback function that takes a location as argument and
      returns a boolean indicating whether to ignore files in that location.
    - error_handler: Error handler callback. No action taken by default.
    """

    if max_depth < 0:
        raise Exception("ERROR: `max_depth` must be a positive integer or 0.")

    # Find root directory depth using path separator's count
    root_dir_depth = root_location.count(os.path.sep)

    for top, dirs, files in os_walk(root_location, topdown=True, onerror=error_handler):
        # If depth is limited (non-zero)
        if max_depth:
            current_depth = top.count(os.path.sep) - root_dir_depth

        if skip_ignored(top) or (max_depth and current_depth >= max_depth):
            # we clear out `dirs` and `files` to prevent `os_walk` from visiting
            # the files and subdirectories of directories we are ignoring or
            # are not in the specified nesting level
            dirs[:] = []
            files[:] = []
            continue
        yield top, dirs, files


@attr.s(slots=True)
class Header(object):
    """
    Represent a Codebase header. Each tool that transforms the codebase
    should create a Header and append it to the Codebase.headers list.
    """

    tool_name = String(help='Name of the tool used such as scancode-toolkit.')
    tool_version = String(default='', help='Tool version used such as v1.2.3.')
    options = Mapping(help='Mapping of key/values describing the options used with this tool.')
    notice = String(default='', help='Notice text for this tool.')
    start_timestamp = String(help='Start timestamp for this header.')
    end_timestamp = String(help='End timestamp for this header.')
    output_format_version = String(help='Version for the output data format, such as v1.1 .')
    duration = String(help='Scan duration in seconds.')
    message = String(help='Message text.')
    errors = List(help='List of error messages.')
    warnings = List(help='List of warning messages.')
    extra_data = Mapping(help='Mapping of extra key/values for this tool.')

    def to_dict(self):
        return attr.asdict(self, dict_factory=dict)

    @classmethod
    def from_dict(cls, **kwargs):
        """
        Return a Header object deserialized from a `kwargs` mapping of
        key/values. Unknown attributes are ignored.
        """
        known_attributes = set(attr.fields_dict(Header))
        kwargs = {k: v for k, v in kwargs.items() if k in known_attributes}
        return cls(**kwargs)


def ignore_nothing(resource, codebase):
    """
    Return True if `resource` should be ignored.

    This function is used as a callable for `ignored` argument in Codebase and
    Resource walk.
    """
    return False


class Codebase:
    """
    Represent a codebase being scanned. A Codebase is a list of Resources.
    """

    # we do not really need slots but this is a way to ensure we have tight
    # control on object attributes
    __slots__ = (
        'max_depth',
        'location',
        'has_single_resource',
        'resource_attributes',
        'resource_class',
        'root',
        'is_file',
        'temp_dir',
        'resources_by_path',
        'resources_count',
        'paths',
        'max_in_memory',
        'all_in_memory',
        'all_on_disk',
        'cache_dir',
        'headers',
        'current_header',
        'codebase_attributes',
        'attributes',
        'counters',
        'timings',
        'errors',
    )

    # the value returned if the resource is cached
    CACHED_RESOURCE = 1

    def __init__(
        self,
        location,
        resource_attributes=None,
        codebase_attributes=None,
        temp_dir=temp_dir,
        max_in_memory=10000,
        max_depth=0,
        paths=tuple(),
        *args,
        **kwargs,
    ):
        """
        Initialize a new codebase rooted at the ``location`` existing file or
        directory.

        Use an optional list of ``paths`` strings that are paths relative to the
        root ``location`` such that joining the root ``location`` and such a
        path is the ``location`` of this path. If these ``paths`` are provided,
        the codebase will only contain these paths and no other path.

        ``resource_attributes`` is an ordered mapping of attr Resource attributes
        such as plugin-provided attributes: these will be added to a Resource
        sub-class crafted for this codebase.

        ``codebase_attributes`` is an ordered mapping of attr Codebase attributes
        such as plugin-provided attributes: these will be added to a
        CodebaseAttributes sub-class crafted for this codebase.

        ``temp_dir`` is the base temporary directory to use to cache resources on
        disk and other temporary files.

        ``max_in_memory`` is the maximum number of Resource instances to keep in
        memory. Beyond this number, Resource are saved on disk instead. -1 means
        no memory is used and 0 means unlimited memory is used.

        ``max_depth`` is the maximum depth of subdirectories to descend below and
        including `location`.
        """
        self.max_depth = max_depth

        # Resource sub-class to use: Configured with attributes in _populate
        self.resource_class = Resource

        self.resource_attributes = resource_attributes or {}
        self.codebase_attributes = codebase_attributes or {}

        # setup location
        ########################################################################
        location = os.fsdecode(location)
        location = abspath(normpath(expanduser(location)))
        location = location.rstrip('/\\')
        # TODO: what if is_special(location)???
        assert exists(location)
        self.location = location

        self.is_file = filetype_is_file(location)

        # True if this codebase root is a file or an empty directory.
        self.has_single_resource = bool(self.is_file or not os.listdir(location))

        ########################################################################
        # Set up caching, summary, timing, and error info
        self._setup_essentials(temp_dir, max_in_memory)

        # finally populate
        self.paths = self._prepare_clean_paths(paths)
        self._populate()

    def _prepare_clean_paths(self, paths=tuple()):
        """
        Return a new set of cleaned ``paths`` possibly empty.
        We convert to POSIX and ensure we have no slash at both ends.
        """
        paths = (clean_path(p) for p in (paths or []) if p)
        # we sort by path segments (e.g. essentially a topo sort)
        _sorter = lambda p: p.split('/')
        return sorted(paths, key=_sorter)

    def _setup_essentials(self, temp_dir=temp_dir, max_in_memory=10000):
        """
        Set the remaining Code=base attributes

        `temp_dir` is the base temporary directory to use to cache resources on
        disk and other temporary files.

        `max_in_memory` is the maximum number of Resource instances to keep in
        memory. Beyond this number, Resource are saved on disk instead. -1 means
        no memory is used and 0 means unlimited memory is used.
        """

        # setup Resources
        ########################################################################
        # root resource, never cached on disk
        self.root = None

        # mapping of {path: Resource}. This the key data structure of a Codebase.
        # All resources MUST exist there. When cached to disk the value is CACHED_RESOURCE
        self.resources_by_path = {}
        self.resources_count = 0

        # setup caching
        ########################################################################
        # dir used for caching and other temp files
        self.temp_dir = temp_dir

        # maximum number of Resource objects kept in memory cached in this
        # Codebase. When the number of in-memory Resources exceed this number,
        # the next Resource instances are saved to disk instead and re-loaded
        # from disk when used/needed.
        self.max_in_memory = max_in_memory
        # use only memory
        self.all_in_memory = max_in_memory == 0
        # use only disk
        self.all_on_disk = max_in_memory == -1

        # dir where the on-disk cache is stored
        self.cache_dir = None
        if not self.all_in_memory:
            # this is unique to this codebase instance
            self.cache_dir = get_codebase_cache_dir(temp_dir=temp_dir)

        # setup extra and misc attributes
        ########################################################################

        # stores a list of Header records for this codebase
        self.headers = []
        self.current_header = None

        # mapping of scan counters at the codebase level such
        # as the number of files and directories, etc
        self.counters = dict()

        # mapping of timings for scan stage as {stage: time in seconds as float}
        # This is populated automatically.
        self.timings = dict()

        # list of error strings from collecting the codebase details (such as
        # unreadable file, etc).
        self.errors = []

    def _get_resource_cache_location(self, path, create_dirs=False):
        """
        Return the location where to get/put a Resource in the cache given a
        Resource `path`. Create the directories if requested.
        """
        if not self.cache_dir:
            return

        if isinstance(path, Resource):
            path = path.path

        path = clean_path(path)

        # for the cached file name, we use an md5 of the path to avoid things being too long
        resid = str(md5(path.encode('utf-8')).hexdigest())
        cache_sub_dir, cache_file_name = resid[-2:], resid

        parent = join(self.cache_dir, cache_sub_dir)
        if create_dirs and not exists(parent):
            create_dir(parent)

        return join(parent, cache_file_name)

    def _collect_codebase_attributes(self, *args, **kwargs):
        """
        Return a mapping of CodebaseAttributes fields to use with this Codebase
        """
        return self.codebase_attributes

    def _build_resource_class(self, *args, **kwargs):
        """
        Return a Resource class to use with this Codebase
        """
        # Resource sub-class to use. Configured with plugin attributes if present
        return attr.make_class(
            name='ScannedResource',
            attrs=self.resource_attributes or {},
            slots=True,
            bases=(Resource,),
        )

    # TODO: add populate progress manager!!!
    def _populate(self):
        """
        Populate this codebase with Resource objects.

        The actual subclass of Resource objects used in this codebase will be
        created as a side effect.

        Population is done by walking its `location` topdown, breadth-first,
        first creating first file then directory Resources both sorted in case-
        insensitive name order.

        Special files, links and VCS files are ignored.
        """
        # Collect headers
        ##########################################################
        self.headers = []

        # Collect codebase-level attributes and build a class, then load
        ##########################################################
        # Codebase attributes to use. Configured with plugin attributes if
        # present.
        self.codebase_attributes = self._collect_codebase_attributes()
        cbac = _CodebaseAttributes.from_attributes(attributes=self.codebase_attributes)
        self.attributes = cbac()

        # Resource sub-class to use. Configured with plugin attributes if present
        ##########################################################
        self.resource_class = self._build_resource_class()

        ##########################################################
        # walk and create resources proper

        # Create root first
        ##########################################################
        root = self._create_root_resource()
        if TRACE:
            logger_debug('Codebase.populate: root:', root)

        if self.has_single_resource:
            # there is nothing else to do for a single file or a single
            # childless directory
            return

        if self.paths:
            return self._create_resources_from_paths(root=root, paths=self.paths)
        else:
            return self._create_resources_from_root(root=root)

    def _create_resources_from_paths(self, root, paths):
        # without paths we iterate the provided paths. We report an error
        # if a path is missing on disk.

        # !!!NOTE: WE DO NOT skip_ignored in this case!!!!!

        base_location = parent_directory(root.location)

        # track resources parents by path during construction to avoid
        # recreating all ancestor directories
        parents_by_path = {root.path: root}

        for path in paths:
            res_loc = join(base_location, path)
            if not exists(res_loc):
                msg = f'ERROR: cannot populate codebase: path: {path!r} not found in {res_loc!r}'
                self.errors.append(msg)
                raise Exception(path, join(base_location, path))
                continue

            # create all parents. The last parent is the one we want to use
            parent = root
            if TRACE:
                logger_debug('Codebase._create_resources_from_paths: parent', parent)
            for parent_path in get_ancestor_paths(path, include_self=False):
                if TRACE:
                    logger_debug(
                        f'  Codebase._create_resources_from_paths: parent_path: {parent_path!r}'
                    )
                if not parent_path:
                    continue
                newpar = parents_by_path.get(parent_path)
                if TRACE:
                    logger_debug('  Codebase._create_resources_from_paths: newpar', repr(newpar))

                if not newpar:
                    newpar = self._get_or_create_resource(
                        name=file_name(parent_path),
                        parent=parent,
                        path=parent_path,
                        is_file=False,
                    )
                    if not newpar:
                        raise Exception(
                            f'ERROR: Codebase._create_resources_from_paths: cannot create parent for: {parent_path!r}'
                        )
                    parent = newpar

                    parents_by_path[parent_path] = parent

                    if TRACE:
                        logger_debug(
                            f'  Codebase._create_resources_from_paths:',
                            f'created newpar: {newpar!r}',
                        )

            res = self._get_or_create_resource(
                name=file_name(path),
                parent=parent,
                path=path,
                is_file=isfile(res_loc),
            )
            if TRACE:
                logger_debug('Codebase._create_resources_from_paths: resource', res)

    def _create_resources_from_root(self, root):
        # without paths we walks the root location top-down

        # track resources parents by location during construction.
        # NOTE: this cannot exhaust memory on a large codebase, because we do
        # not keep parents already walked and we walk topdown.
        parents_by_loc = {root.location: root}

        def err(_error):
            """os.walk error handler"""
            self.errors.append(
                f'ERROR: cannot populate codebase: {_error}\n{traceback.format_exc()}'
            )

        # Walk over the directory and build the resource tree
        for (top, dirs, files) in depth_walk(
            root_location=root.location,
            max_depth=self.max_depth,
            error_handler=err,
        ):
            parent = parents_by_loc.pop(top)
            for created in self._create_resources(
                parent=parent,
                top=top,
                dirs=dirs,
                files=files,
            ):
                # on the plain, bare FS, files cannot be parents
                if not created.is_file:
                    parents_by_loc[created.location] = created

    def _create_resources(self, parent, top, dirs, files, skip_ignored=skip_ignored):
        """
        Create and yield ``files`` and ``dirs`` children Resources of a
        ``parent`` Resource. These are sorted as: directories then files and by
        lowercase name, then name.
        """
        for names, is_file in [(dirs, False), (files, True)]:
            names.sort(key=lambda p: (p.lower(), p))

            for name in names:
                location = join(top, name)
                if skip_ignored(location):
                    continue
                res = self._get_or_create_resource(
                    name=name,
                    parent=parent,
                    is_file=is_file,
                )
                if TRACE:
                    logger_debug('Codebase.create_resources:', res)
                yield res

    def _create_root_resource(self):
        """
        Create and return the root Resource of this codebase.
        """
        # we cannot recreate a root if it exists!!
        if self.root:
            raise TypeError('Root resource already exists and cannot be recreated')

        location = self.location
        name = file_name(location)

        # do not strip root for codebase with a single Resource.
        path = Resource.build_path(root_location=location, location=location)

        if TRACE:
            logger_debug(f'  Codebase._create_root_resource: {path} is_file: {self.is_file}')
            logger_debug()

        root = self.resource_class(
            name=name,
            location=location,
            # never cached
            cache_location=None,
            path=path,
            is_root=True,
            is_file=self.is_file,
        )

        self.resources_by_path[path] = root
        self.resources_count += 1
        self.root = root
        return root

    def _get_or_create_resource(
        self,
        name,
        parent,
        is_file=False,
        path=None,
    ):
        """
        Create and return a new codebase Resource with ``path`` and ``location``.
        """
        if not parent:
            raise TypeError(
                f'Cannot create resource without parent: name: {name!r}, path: {path!r}'
            )

        # If the codebase is virtual, we provide the path
        if not path:
            path = posixpath_join(parent.path, name)
        path = clean_path(path)

        existing = self.get_resource(path)
        if existing:
            if TRACE:
                logger_debug('  Codebase._get_or_create_resource: path  already exists:', path)
            return existing

        if self._use_disk_cache_for_resource():
            cache_location = self._get_resource_cache_location(path=path, create_dirs=True)
        else:
            cache_location = None

        # NOTE: If the codebase is virtual, then there is no location
        parent_location = parent.location
        if parent_location:
            location = join(parent_location, name)
        else:
            location = None

        if TRACE:
            logger_debug(
                f'  Codebase._get_or_create_resource: with path: {path}\n'
                f'  name={name}, is_file={is_file}'
            )

        child = self.resource_class(
            name=name,
            location=location,
            path=path,
            cache_location=cache_location,
            is_file=is_file,
        )
        self.resources_count += 1

        parent.children_names.append(name)
        self.save_resource(parent)
        self.save_resource(child)
        return child

    def get_or_create_current_header(self):
        """
        Return the current Header. Create it if it does not exists and store
        it in the headers.
        """
        if not self.current_header:
            self.current_header = Header()
            self.headers.append(self.current_header)
        return self.current_header

    def get_files_count(self):
        """
        Return the final files counts for the codebase.
        """
        return self.counters.get('final:files_count', 0)

    def add_files_count_to_current_header(self):
        """
        Add the final files counts for the codebase to the current header.
        Return the files_count.
        """
        files_count = self.get_files_count()
        current_header = self.get_or_create_current_header()
        current_header.extra_data['files_count'] = files_count
        return files_count

    def get_headers(self):
        """
        Return a serialized headers composed only of native Python objects
        suitable for use in outputs.
        """
        return [le.to_dict() for le in (self.headers or [])]

    def exists(self, resource):
        """
        Return True if the Resource path exists in the codebase.
        """
        return resource and resource.path in self.resources_by_path

    def _use_disk_cache_for_resource(self):
        """
        Return True if Resource ``res`` should be cached on-disk or False if it
        should be kept in-memory.
        """

        use_disk_cache = False
        if self.all_on_disk:
            use_disk_cache = True
        elif self.all_in_memory:
            use_disk_cache = False
        else:
            # mixed case where some are in memory and some on disk
            if self.resources_count < self.max_in_memory:
                use_disk_cache = False
            else:
                use_disk_cache = True

        if TRACE:
            logger_debug(
                f'    Codebase._use_disk_cache_for_resource mode: {use_disk_cache} '
                f'on_disk: {self.all_on_disk} '
                f'in_mem: {self.all_in_memory} '
                f'max_in_mem: {self.max_in_memory}'
            )
        return use_disk_cache

    def _exists_in_memory(self, path):
        """
        Return True if Resource `path` exists in the codebase memory cache.
        """
        path = clean_path(path)
        return isinstance(self.resources_by_path.get(path), Resource)

    def _exists_on_disk(self, path):
        """
        Return True if Resource `path` exists in the codebase disk cache.
        """
        path = clean_path(path)
        if not self._exists_in_memory(path):
            cache_location = self._get_resource_cache_location(path, create_dirs=False)
            if cache_location:
                return exists(cache_location)

    ########### FIXME: the PATH SHOULD NOT INCLUDE THE ROOT NAME
    def get_resource(self, path):
        """
        Return the Resource with `path` or None if it does not exists.
        The ``path`` must be relative to the root (and including the root
        name as its first segment).
        """
        assert isinstance(path, str), f'Invalid path: {path!r} is not a string.'
        path = clean_path(path)
        if TRACE:
            msg = ['  Codebase.get_resource:', 'path:', path]
            if not path or path not in self.resources_by_path:
                msg.append('not in resources!')
            else:
                msg.extend(['exists on disk:', self._exists_on_disk(path)])
                msg.extend(['exists in memo:', self._exists_in_memory(path)])
            logger_debug(*msg)

        # we use Codebase.CACHED_RESOURCE as a semaphore for existing but only
        # on-disk, non-in-memory resource that we need to load from the disk
        # cache to differentiate from None which means missing
        res = self.resources_by_path.get(path)
        if res is Codebase.CACHED_RESOURCE:
            res = self._load_resource(path)

        elif isinstance(res, Resource):
            res = attr.evolve(res)

        elif res is None:
            pass
        else:
            # this should never happen
            raise Exception(f'get_resource: Internal error when getting {path!r}')

        if TRACE:
            logger_debug('    Resource:', res)
        return res

    def save_resource(self, resource):
        """
        Save the `resource` Resource to cache (in memory or disk).
        """
        if not resource:
            return

        path = clean_path(resource.path)

        if TRACE:
            logger_debug('  Codebase.save_resource:', resource)

        if resource.is_root:
            self.root = resource
            self.resources_by_path[path] = resource

        elif resource.cache_location:
            self._dump_resource(resource)
            self.resources_by_path[path] = Codebase.CACHED_RESOURCE

        else:
            self.resources_by_path[path] = resource

    def _dump_resource(self, resource):
        """
        Dump a Resource to the disk cache.
        """
        cache_location = resource.cache_location

        if not cache_location:
            raise TypeError(
                'Resource cannot be dumped to disk and is used only' f'in memory: {resource}'
            )

        # TODO: consider messagepack or protobuf for compact/faster processing?
        with open(cache_location, 'w') as cached:
            cached.write(json.dumps(resource.serialize(), check_circular=False))

    # TODO: consider adding a small LRU cache in front of this for perf?
    def _load_resource(self, path):
        """
        Return a Resource with ``path`` loaded from the disk cache.
        """
        path = clean_path(path)
        cache_location = self._get_resource_cache_location(path, create_dirs=False)

        if TRACE:
            logger_debug(
                '    Codebase._load_resource: exists:',
                exists(cache_location),
                'cache_location:',
                cache_location,
            )

        if not exists(cache_location):
            raise ResourceNotInCache(f'Failed to load Resource: {path} from {cache_location!r}')

        # TODO: consider messagepack or protobuf for compact/faster processing
        try:
            with open(cache_location, 'rb') as cached:
                # TODO: Use custom json encoder to encode JSON list as a tuple
                # TODO: Consider using simplejson
                data = json.load(cached)
                return self.resource_class(**data)
        except Exception as e:
            with open(cache_location, 'rb') as cached:
                cached_data = cached.read()
            msg = (
                f'ERROR: failed to load resource from cached location: {cache_location} '
                'with content:\n\n' + repr(cached_data) + '\n\n' + traceback.format_exc()
            )
            raise Exception(msg) from e

    def _remove_resource(self, resource):
        """
        Remove the ``resource`` Resource object from this codebase.
        Does not remove children.
        """
        if resource.is_root:
            raise TypeError(f'Cannot remove the root resource from codebase: {resource!r}')

        # remove from in-memory cache. The disk cache is cleared on exit.
        self.resources_by_path.pop(resource.path, None)
        if TRACE:
            logger_debug('Codebase._remove_resource:', resource)

    def remove_resource(self, resource):
        """
        Remove the `resource` Resource object and all its children from the
        codebase. Return a set of removed Resource paths.
        """
        if TRACE:
            logger_debug('Codebase.remove_resource')
            logger_debug('  resource', resource)

        if resource.is_root:
            raise TypeError(f'Cannot remove the root resource from codebase: {resource!r}')

        removed_paths = set()

        # remove all descendants bottom up to avoid out-of-order access to
        # removed resources
        for descendant in resource.walk(self, topdown=False):
            self._remove_resource(descendant)
            removed_paths.add(descendant.location)

        # remove resource from parent
        parent = resource.parent(self)
        if TRACE:
            logger_debug('    parent', parent)
        parent.children_names.remove(resource.name)
        parent.save(self)

        # remove resource proper
        self._remove_resource(resource)
        removed_paths.add(resource.location)

        return removed_paths

    def walk(self, topdown=True, skip_root=False, ignored=ignore_nothing):
        """
        Yield all resources for this Codebase walking its resource tree. Walk
        the tree top-down, depth-first if ``topdown`` is True, otherwise walk
        bottom-up.

        Each level is sorted by children woth this sort order: resource without-
        children first, then resource with-children and each group sorted by
        case-insensitive name.

        If ``skip_root`` is True, the root resource is not returned unless this
        is a codebase with a single resource.

        ``ignored`` is a callable that accepts two arguments, ``resource`` and
        ``codebase``, and returns True if ``resource`` should be ignored.
        """
        root = self.root

        if ignored(resource=root, codebase=self):
            return

        # make a copy
        root = attr.evolve(root)

        # include root if no children (e.g. codebase with a single resource)
        if self.has_single_resource or (skip_root and not root.has_children()):
            skip_root = False

        root = attr.evolve(root)
        if topdown and not skip_root:
            yield root

        for res in root.walk(self, topdown=topdown, ignored=ignored):
            yield res

        if not topdown and not skip_root:
            yield root

    def __iter__(self):
        yield from self.walk()

    def walk_filtered(self, topdown=True, skip_root=False):
        """
        Walk this Codebase as with walk() but does not return Resources with
        `is_filtered` flag set to True.
        """
        for resource in self.walk(topdown=topdown, skip_root=skip_root):
            if not resource.is_filtered:
                yield resource

    def compute_counts(self, skip_root=False, skip_filtered=False):
        """
        Compute, update and save the counts of every resource.
        Return a tuple of top level counters for this codebase as:
          (files_count, dirs_count, size_count).

        The counts are computed differently based on these flags:
        - If ``skip_root`` is True, the root resource is not included in counts.
        - If ``skip_filtered`` is True, resources with ``is_filtered`` set to True
          are not included in counts.
        """
        self.update_counts(skip_filtered=skip_filtered)

        root = self.root
        files_count = root.files_count
        dirs_count = root.dirs_count
        size_count = root.size_count

        if (skip_root and not root.is_file) or (skip_filtered and root.is_filtered):
            return files_count, dirs_count, size_count

        if root.is_file:
            files_count += 1
        else:
            dirs_count += 1
        size_count += root.size or 0

        return files_count, dirs_count, size_count

    def update_counts(self, skip_filtered=False):
        """
        Update files_count, dirs_count and size_count attributes of each
        Resource in this codebase based on the current Resource data.

        If ``skip_filtered`` is True, resources with ``is_filtered`` set to True are
        not included in counts.
        """
        # note: we walk bottom up to update things in the proper order
        # and the walk MUST NOT skip filtered, only the compute
        for resource in self.walk(topdown=False):
            try:
                resource._compute_children_counts(self, skip_filtered)
            except Exception as e:
                msg = f'ERROR: cannot compute children counts for: {resource.path}'
                raise Exception(msg) from e

    def clear(self):
        """
        Purge the codebase cache(s).
        """
        delete(self.cache_dir)

    def lowest_common_parent(self):
        """
        Return a Resource that is the lowest common parent (aka. lowest common
        ancestor) of all the files of this codebase, skipping root directory
        segments that are "empty" e.g. with a single child. Return None is this
        codebase contains a single resource.
        """
        if self.has_single_resource:
            return self.root

        for res in self.walk(topdown=True):
            if not res.is_file:
                kids = res.children(self)
                if len(kids) == 1 and not kids[0].is_file:
                    # this is an empty dir with a single dir child, therefore
                    # we shall continue the descent walk
                    continue
                else:
                    # the dir starts to branch: we have our lowest common parent
                    # root
                    break
            else:
                # we are in a case that should never happen
                return self.root
        return res

    def to_list(
        self,
        with_timing=False,
        with_info=False,
        skinny=False,
        full_root=False,
        strip_root=False,
    ):
        """
        Return a list of all Resources of this Codebase as mappings.
        """
        if self.has_single_resource:
            return [
                self.root.to_dict(
                    with_timing=with_timing,
                    with_info=with_info,
                    skinny=skinny,
                    # we never strip root for single res codebase
                    full_root=full_root,
                    strip_root=False,
                )
            ]

        td = partial(
            Resource.to_dict,
            with_timing=with_timing,
            with_info=with_info,
            skinny=skinny,
            full_root=full_root,
            strip_root=strip_root,
        )
        return [td(r) for r in self.walk(skip_root=strip_root)]


def to_decoded_posix_path(path):
    """
    Return `path` as a Unicode POSIX path given a unicode or bytes path string.
    """
    return clean_path(os.fsdecode(as_posixpath(path)))


@attr.attributes(slots=True)
class Resource(object):
    """
    A resource represents a file or directory with essential "file information"
    and the scanned data details.

    A Resource is a tree that models the fileystem tree structure.

    In order to support lightweight and smaller objects that can be serialized
    and deserialized (such as pickled in multiprocessing) without pulling in a
    whole object tree, a Resource does not store its related objects directly:
    - the Codebase it belongs to is never stored.
    - its parent Resource and its Resource children objects are queryable by path.

    Querying the Resource relationships and walking the Resources tree typically
    requires to lookup the corresponding object by path in the Codebase object.
    """

    # the file or directory name in the OS preferred representation (either
    # bytes on Linux and Unicode elsewhere)
    name = attr.attrib(repr=False)

    # the file or directory absolute location in the OS preferred representation
    # (either bytes on Linux and Unicode elsewhere) using the OS native path
    # separators.
    location = attr.attrib(repr=False)

    # the file or directory POSIX path decoded as unicode using the filesystem
    # encoding. This is the path that will be reported in output and is always
    # relative to and starting with the root directory.
    path = attr.attrib(converter=to_decoded_posix_path)

    # location of the file where this resource can be chached on disk in the OS
    # preferred representation (either bytes on Linux and Unicode elsewhere)
    cache_location = attr.attrib(default=None, repr=False)

    # True for file, False for directory
    is_file = attr.ib(default=False)

    # True if this Resource should be filtered out, e.g. skipped from the
    # returned list of resources
    is_filtered = attr.ib(default=False)

    # a list of names
    children_names = attr.ib(default=attr.Factory(list), repr=TRACE)

    # external data to serialize
    size = attr.ib(default=0, type=int, repr=TRACE)

    # These attributes are re/computed for directories and files with children
    # they represent are the for the full descendants of a Resource
    size_count = attr.ib(default=0, type=int, repr=False)
    files_count = attr.ib(default=0, type=int, repr=False)
    dirs_count = attr.ib(default=0, type=int, repr=False)

    # list of scan error strings
    scan_errors = attr.ib(default=attr.Factory(list), repr=False)

    # Duration in seconds as float to run all scans for this resource
    scan_time = attr.ib(default=0, repr=False)

    # mapping of timings for each scan as {scan_key: duration in seconds as a float}
    scan_timings = attr.ib(default=attr.Factory(dict), repr=False)

    # stores a mapping of extra data for this Resource this data is never
    # returned in a to_dict() and not meant to be saved in the final scan
    # results. Instead it can be used to store extra data attributes that may be
    # useful during a scan processing but are not usefuol afterwards. Be careful
    # not to override keys/values that may have been created by some other
    # plugin or process
    extra_data = attr.ib(default=attr.Factory(dict), repr=False)

    is_root = attr.ib(default=False, type=bool, repr=False)

    @property
    def type(self):
        return 'file' if self.is_file else 'directory'

    @type.setter
    def type(self, value):
        if value == 'file':
            self.is_file = True
        else:
            self.is_file = False

    @classmethod
    def build_path(cls, root_location, location):
        """
        Return a POSIX path string (using "/"  separators) of ``location`` relative
        to ``root_location`. Both locations are absolute native locations.
        The returned path has no leading and trailing slashes.  The first segment
        of this path is always the last segment of the ``root_location``.
        For example:
        >>> result = Resource.build_path(r'D:\\foo\\bar', r'D:\\foo\\bar\\baz')
        >>> assert result == 'bar/baz', repr(result)
        >>> result = Resource.build_path('/foo/bar/', '/foo/bar/baz')
        >>> assert result == 'bar/baz', result
        >>> result = Resource.build_path('/foo/bar/', '/foo/bar')
        >>> assert result  == 'bar', result
        """
        root_loc = clean_path(root_location)
        loc = clean_path(location)
        assert loc.startswith(root_loc)

        # keep the root directory name by default
        root_loc = posixpath_parent(root_loc).strip('/')
        path = loc.replace(root_loc, '', 1).strip('/')
        if TRACE:
            logger_debug('build_path:', root_loc, loc, path)
        return path

    def get_path(self, full_root=False, strip_root=False):
        """
        Return a POSIX path string (using "/"  separators) for this resource.
        The returned path has no leading and trailing slashes.

        - If ``full_root`` is True, return an absolute path.

        - If ``strip_root`` is True, return a relative path without the first
          root segment. Ignored if ``full_root`` is True.

        - Otherwise return a relative path where the first segment is the
          ``location`` last path segment.
        """
        if full_root:
            return self.full_root_path
        elif strip_root:
            return self.strip_root_path
        else:
            return self.path

    @property
    def full_root_path(self):
        """
        Return a fully rooted POSIX path stripped from leading and trailing slash
        """
        location = self.location
        if location:
            return clean_path(as_posixpath(self.location))
        else:
            return self.path

    @property
    def strip_root_path(self):
        """
        Return a path relative to the root, stripped from its root segment
        unless the codebase contains a single file or there is only one segment
        in the path.
        """
        return strip_first_path_segment(self.path)

    @property
    def is_dir(self):
        # note: we only store is_file
        return not self.is_file

    @property
    def base_name(self):
        # FIXME: we should call the function only once
        base_name, _extension = splitext_name(self.name, is_file=self.is_file)
        return base_name

    @base_name.setter
    def base_name(self, value):
        pass

    @property
    def extension(self):
        # FIXME: we should call the function only once
        _base_name, extension = splitext_name(self.name, is_file=self.is_file)
        return extension

    @extension.setter
    def extension(self, value):
        pass

    def extracted_to(self, codebase):
        """
        Return the path this Resource archive was extracted to or None.
        """
        extract_path = f'{self.path}-extract'
        return codebase.get_resource(extract_path)

    def extracted_from(self, codebase):
        """
        Return the path to an archive this Resource was extracted from or None.
        """
        path = self.path
        if '-extract' in path:
            archive_path, _, _ = self.path.rpartition('-extract')
            return codebase.get_resource(archive_path)

    @classmethod
    def get(cls, codebase, path):
        """
        Return the Resource with `path` in `codebase` or None if it does not
        exists.
        """
        return codebase.get_resource(path)

    def save(self, codebase):
        """
        Save this resource in `codebase` (in memory or disk).
        """
        return codebase.save_resource(self)

    def remove(self, codebase):
        """
        Remove this resource and all its children from the codebase.
        Return a set of removed Resource paths.
        """
        return codebase.remove_resource(self)

    def _compute_children_counts(self, codebase, skip_filtered=False):
        """
        Compute counts and update self with these counts from direct children.
        Return a tuple of counters (files_count, dirs_count, size_count) for the
        direct children of this Resource.

        If `skip_filtered` is True, skip resources with the `is_filtered` flag
        set to True.

        Note: because certain files such as archives can have children, they may
        have a files and dirs counts. The size of a directory is aggregated size
        of its files (including the count of files inside archives).
        """
        files_count = dirs_count = size_count = 0
        for child in self.children(codebase):
            files_count += child.files_count or 0
            dirs_count += child.dirs_count or 0
            size_count += child.size_count or 0

            if skip_filtered and child.is_filtered:
                continue

            if child.is_file:
                files_count += 1
            else:
                dirs_count += 1
            size_count += child.size or 0

        self.files_count = files_count
        self.dirs_count = dirs_count
        self.size_count = size_count
        self.save(codebase)

        return files_count, dirs_count, size_count

    def walk(self, codebase, topdown=True, ignored=ignore_nothing):
        """
        Yield all descendant Resources of this Resource. Does not include self.

        Walk the tree top-down, depth-first if `topdown` is True, otherwise walk
        bottom-up.

        Each level is sorted by children sort order (e.g. without-children, then
        with-children and each group by case-insensitive name)

        `ignored` is a callable that accepts two arguments, `resource` and `codebase`,
        and returns True if `resource` should be ignored.
        """

        for child in self.children(codebase):
            if not ignored(child, codebase):
                child = attr.evolve(child)
                if topdown:
                    yield child

                for subchild in child.walk(
                    codebase=codebase,
                    topdown=topdown,
                    ignored=ignored,
                ):
                    if not ignored(subchild, codebase):
                        yield subchild

                if not topdown:
                    yield child

    def has_children(self):
        """
        Return True is this Resource has children.
        """
        return bool(self.children_names)

    def children(self, codebase, names=()):
        """
        Return a sorted sequence of direct children Resource objects for this
        Resource or an empty sequence.

        Sorting is by resources without children, then resource with children
        (e.g. directories or files with children), then case-insentive name.
        """
        children_names = self.children_names or []
        if not children_names:
            return []

        if names:
            kids = set(children_names)
            children_names = [n for n in names if n in kids]
            if not children_names:
                return []

        child_path = partial(posixpath_join, self.path)
        get_child = codebase.get_resource
        children = [get_child(path=child_path(name)) for name in children_names]

        _sorter = lambda r: (r.has_children(), r.name.lower(), r.name)
        return sorted((c for c in children if c), key=_sorter)

    def has_parent(self):
        """
        Return True is this Resource has children.
        """
        return not self.is_root

    def parent_path(self):
        """
        Return the parent Resource object for this Resource or None.
        """
        return self.has_parent() and parent_directory(self.path, with_trail=False)

    def parent(self, codebase):
        """
        Return the parent Resource object for this Resource or None.
        """
        parent_path = self.parent_path()
        return parent_path and codebase.get_resource(parent_path)

    def has_siblings(self, codebase):
        """
        Return True is this Resource has siblings.
        """
        return self.has_parent() and self.parent(codebase).has_children()

    def siblings(self, codebase):
        """
        Return a sequence of sibling Resource objects for this Resource
        or an empty sequence.
        """
        if self.has_parent():
            return self.parent(codebase).children(codebase)
        return []

    def ancestors(self, codebase):
        """
        Return a sequence of ancestor Resource objects from root to self
        (includes self).
        """
        if self.is_root:
            return [self]

        ancestors = deque()
        ancestors_appendleft = ancestors.appendleft
        current = self

        # walk up the parent tree up to the root
        while current and not current.is_root:
            ancestors_appendleft(current)
            current = current.parent(codebase)

        # append root too
        if current:
            ancestors_appendleft(current)

        return list(ancestors)

    def descendants(self, codebase):
        """
        Return a sequence of descendant Resource objects
        (does NOT include self).
        """
        return list(self.walk(codebase=codebase, topdown=True))

    def distance(self, codebase):
        """
        Return the distance as the number of path segments separating this
        Resource from the `codebase` root Resource.

        The codebase root has a distance of zero ot itself. Its direct children
        have a distance of one, and so on.
        """
        if self.is_root:
            return 0
        return len(self.ancestors(codebase)) - 1

    def to_dict(
        self,
        with_timing=False,
        with_info=False,
        skinny=False,
        full_root=False,
        strip_root=False,
    ):
        """
        Return a mapping of representing this Resource and its data.

        The path is always a POSIX path stripped from leading and trailing
        slashes and can be either one of these exclusive flags:

        - If ``full_root`` is True, this is a full path when available.

        - If ``strip_root`` is True, this is a path relative to the root,
          stripped from its root segment unless the codebase contains a single
          file with a single root segment. Ignored if ``full_root`` is True.
        """
        path = self.get_path(full_root=full_root, strip_root=strip_root)
        res = dict(path=path, type=self.type)
        if skinny:
            return res

        if with_info:
            res['name'] = self.name
            res['base_name'] = self.base_name
            res['extension'] = self.extension
            res['size'] = self.size

        # exclude by default all of the "standard", default Resource fields
        self_fields_filter = attr.filters.exclude(*attr.fields(Resource))

        # this will catch every attribute that has been added dynamically, such
        # as scan-provided resource_attributes
        other_data = attr.asdict(self, filter=self_fields_filter, dict_factory=dict)

        # FIXME: make a deep copy of the data first!!!!
        # see https://github.com/nexB/scancode-toolkit/issues/1199
        res.update(other_data)

        if with_timing:
            res['scan_time'] = self.scan_time or 0
            res['scan_timings'] = self.scan_timings or dict()

        if with_info:
            res['files_count'] = self.files_count
            res['dirs_count'] = self.dirs_count
            res['size_count'] = self.size_count

        res['scan_errors'] = self.scan_errors
        if TRACE:
            logger_debug('Resource.to_dict:', res)
        return res

    def serialize(self):
        """
        Return a mapping of representing this Resource and its data in a form
        that is fully serializable (to JSON, YAML, pickle, etc.) and can be used
        to reconstruct a Resource.
        """
        # we save all fields, not just the one in .to_dict()
        serializable = attr.asdict(self)
        serializable['name'] = self.name
        if self.location:
            serializable['location'] = self.location
        if self.cache_location:
            serializable['cache_location'] = self.cache_location
        return serializable


def clean_path(path):
    """
    Return a cleaned and normalized POSIX ``path``.
    """
    path = path or ''
    # convert to posix and ensure we have no slash at both ends
    path = posixpath_normpath(path.replace('\\', '/').strip('/'))
    if path == '.':
        path = ''
    return path


def strip_first_path_segment(path):
    """
    Return a POSIX ``path`` stripped from its first path segment unless there is
    only one segment in which case we return this segment. The returned path has
    no leading and trailing slashes.

    For example::
        >>> strip_first_path_segment('')
        ''
        >>> strip_first_path_segment('foo')
        ''
        >>> strip_first_path_segment('foo/bar/baz')
        'bar/baz'
        >>> strip_first_path_segment('/foo/bar/baz/')
        'bar/baz'
        >>> strip_first_path_segment('foo/')
        ''
    """
    path = clean_path(path)
    if '/' in path:
        _root, _, path = path.partition('/')
        return path
    else:
        return ''


def get_codebase_cache_dir(temp_dir):
    """
    Return a new, created and unique per-run cache storage directory path rooted
    at the `temp_dir` base temp directory in the OS-preferred representation
    (either bytes on Linux and Unicode elsewhere).
    """
    from commoncode.fileutils import get_temp_dir
    from commoncode.timeutils import time2tstamp

    prefix = 'scancode-codebase-' + time2tstamp() + '-'
    return get_temp_dir(base_dir=temp_dir, prefix=prefix)


@attr.s(slots=True)
class _CodebaseAttributes(object):

    def to_dict(self):
        return attr.asdict(self, dict_factory=dict)

    @classmethod
    def from_attributes(cls, attributes):
        """
        Return a new sub class of _CodebaseAttributes built with the
        ``attributes`` mapping of "attr" attributes.
        """
        return attr.make_class(
            name='CodebaseAttributes',
            attrs=attributes or {},
            slots=True,
            bases=(_CodebaseAttributes,),
        )


def build_attributes_defs(mapping, ignored_keys=()):
    """
    Given a mapping, return an ordered mapping of attributes built from the
    mapping keys and values.
    """
    attributes = {}

    # We add the attributes that are not in standard_res_attributes already
    # FIXME: we should not have to infer the schema may be?
    for key, value in mapping.items():
        if key in ignored_keys or key in attributes:
            continue
        if isinstance(value, (list, tuple)):
            attributes[key] = attr.ib(default=attr.Factory(list), repr=False)
        elif isinstance(value, dict):
            attributes[key] = attr.ib(default=attr.Factory(dict), repr=False)
        elif isinstance(value, bool):
            attributes[key] = attr.ib(default=False, type=bool, repr=False)
        elif isinstance(value, int):
            attributes[key] = attr.ib(default=0, type=bool, repr=False)
        else:
            attributes[key] = attr.ib(default=None, repr=False)

    return attributes


class VirtualCodebase(Codebase):

    __slots__ = (
        # TRUE iff the loaded virtual codebase has file information
        'with_info',
        'has_single_resource',
    )

    def __init__(
        self,
        location,
        resource_attributes=None,
        codebase_attributes=None,
        temp_dir=temp_dir,
        max_in_memory=10000,
        paths=tuple(),
        *args,
        **kwargs,
    ):
        """
        Initialize a new virtual codebase from JSON scan file at `location`.
        See the Codebase parent class for other arguments.

        `max_depth`, if passed, will be ignored as VirtualCodebase will
        be using the depth of the original scan.
        """
        logger_debug(f'VirtualCodebase: new from: {location!r}')

        self._setup_essentials(temp_dir, max_in_memory)

        self.codebase_attributes = codebase_attributes or {}
        self.resource_attributes = resource_attributes or {}
        self.resource_class = None
        self.has_single_resource = False
        self.location = location

        scan_data = self._get_scan_data(location)
        self.paths = self._prepare_clean_paths(paths)
        self._populate(scan_data)

    def _get_scan_data_helper(self, location):
        """
        Return scan data loaded from `location`, which is a path string
        """
        try:
            return json.loads(location)
        except:

            location = abspath(normpath(expanduser(location)))
            with open(location) as f:
                scan_data = json.load(f)
            return scan_data

    def _get_scan_data(self, location):
        """
        Return scan data loaded from `location` that is either:
        - a path string
        - a JSON string
        - a Python mapping
        - a List or Tuple of paths to JSON scans to combine together. In this
          case all paths are prefixed with codebase-1/, codebase-2., etc.
          incremented for each location.
        Loading also cleans the paths as POSIX.
        """
        if isinstance(location, dict):
            return location

        if isinstance(
            location,
            (
                list,
                tuple,
            ),
        ):
            combined_scan_data = dict(headers=[], files=[])
            for idx, loc in enumerate(location, 1):
                scan_data = self._get_scan_data_helper(loc)
                headers = scan_data.get('headers')
                if headers:
                    combined_scan_data['headers'].extend(headers)
                files = scan_data.get('files')
                if files:
                    for f in files:
                        f['path'] = posixpath_join(f'codebase-{idx}', clean_path(f['path']))
                    combined_scan_data['files'].extend(files)
                else:
                    raise Exception(
                        f'Input file is missing a "files" (aka. resources) section to load: {loc}'
                    )

            combined_scan_data['headers'] = sorted(
                combined_scan_data['headers'],
                key=lambda x: x['start_timestamp'],
            )
            return combined_scan_data

        return self._get_scan_data_helper(location)

    def _create_empty_resource_data(self):
        """
        Return a dictionary of Resource fields and their default values.

        The fields returned are that which are not part of the standard set of
        Resource attributes.
        """
        # Get fields from the base Resource class and the ScannedResource class
        base_fields = attr.fields(Resource)
        resource_fields = attr.fields(self.resource_class)
        # A dict of {field: field_default_value} for the dynamically created fields
        resource_data = {}
        for field in resource_fields:
            if field in base_fields:
                # We only want the fields that are not part of the base set of fields
                continue
            value = field.default
            if isinstance(value, attr.Factory):
                # For fields that have Factories as values, we set their values
                # to be an instance of whatever type the factory makes
                value = value.factory()
            resource_data[field.name] = value
        return resource_data

    def _collect_codebase_attributes(self, scan_data, *args, **kwargs):
        """
        Return a mapping of CodebaseAttributes fields to use with this Codebase
        """
        # collect attributes from scan data
        all_attributes = (
            build_attributes_defs(
                mapping=scan_data,
                ignored_keys=('headers', 'files'),
            )
            or {}
        )

        # We add in the attributes that we collected from the plugins. They come
        # last for now.
        for name, plugin_attribute in self.codebase_attributes.items():
            if name not in all_attributes:
                all_attributes[name] = plugin_attribute

        return all_attributes

    def _build_resource_class(self, sample_resource_data, *args, **kwargs):
        """
        Return a Resource class to use with this Codebase
        """
        # Collect the existing attributes of the standard Resource class
        standard_res_attributes = set(f.name for f in attr.fields(Resource))

        # add these properties since they are fields but are serialized
        properties = set(['type', 'base_name', 'extension'])
        standard_res_attributes.update(properties)

        # We collect attributes that are not in standard_res_attributes already
        # FIXME: we should not have to infer the schema may be?
        all_res_attributes = build_attributes_defs(
            mapping=sample_resource_data,
            ignored_keys=standard_res_attributes,
        )

        # We add the attributes that we collected from the plugins. They come
        # last for now.
        for name, plugin_attribute in self.resource_attributes.items():
            if name not in all_res_attributes:
                all_res_attributes[name] = plugin_attribute

        # Create the Resource class with the desired attributes
        return attr.make_class(
            name='ScannedResource',
            attrs=all_res_attributes or dict(),
            slots=True,
            bases=(Resource,),
        )

    def _populate(self, scan_data):
        """
        Populate this codebase with Resource objects.

        The actual subclass of Resource objects used in this codebase will be
        created as a side effect.

        Population is done by loading JSON scan results and creating new
        Resources for each files mappings.
        """
        # Collect headers
        ##########################################################
        headers = scan_data.get('headers') or []
        headers = [Header.from_dict(**hle) for hle in headers]
        self.headers = headers

        # Collect codebase-level attributes and build a class, then load
        ##########################################################
        # Codebase attributes to use. Configured with scan_data and plugin
        # attributes if present.
        self.codebase_attributes = self._collect_codebase_attributes(scan_data)
        cbac = _CodebaseAttributes.from_attributes(attributes=self.codebase_attributes)
        self.attributes = cbac()

        # now populate top level codebase attributes
        ##########################################################
        for attr_name in self.codebase_attributes:
            value = scan_data.get(attr_name)
            setattr(self.attributes, attr_name, value)

        ##########################################################
        files_data = scan_data.get('files')
        if not files_data:
            raise Exception('Input has no "files" top-level scan results.')

        if len(files_data) == 1:
            # we will shortcut to populate the codebase with a single root resource
            self.has_single_resource = True
            root_is_file = files_data[0].get('type') == 'file'
        else:
            root_is_file = False

        # Create a virtual root if we are merging multiple input scans together
        location = self.location
        multiple_inputs = (
            isinstance(
                location,
                (
                    list,
                    tuple,
                ),
            )
            and len(location) > 1
        )

        # Iterate through all Resources to collect any attribute in any resource
        # as sample data. The paths were cleaned on loading
        # NOTE: We also:
        # - add a new "segments" attributes with path split in segments
        # - populate a set of unique root names to to check if all scanned
        #   Resources share a common root or need a new virtual root

        root_names = set()
        root_names_add = root_names.add

        sample_resource_data = {}
        sample_resource_data_update = sample_resource_data.update

        for fdata in files_data:
            sample_resource_data_update(fdata)
            segments = fdata['path'].split('/')
            root_names_add(segments[0])
            fdata['path_segments'] = segments

        # Resource sub-class to use. Configured with all known scanned file
        # attributes and plugin attributes if present
        ##########################################################
        self.resource_class = self._build_resource_class(sample_resource_data)

        # do we have file information attributes in this codebase data?
        self.with_info = any(
            a in sample_resource_data
            for a in (
                'name',
                'base_name',
                'extension',
                'size',
                'files_count',
                'dirs_count',
                'size_count',
            )
        )

        # walk and create resources proper
        # Create root resource first
        ##########################################################
        if not root_names:
            raise Exception('Unable to find root for codebase.')

        len_root_names = len(root_names)
        if len_root_names == 1:
            root_path = root_names.pop()
            needs_new_virtual_root = False
        elif len_root_names > 1 or multiple_inputs:
            root_path = 'virtual_root'
            needs_new_virtual_root = True

        if needs_new_virtual_root:
            for fdata in files_data:
                rpath = fdata['path']
                fdata['path'] = posixpath_join(root_path, rpath)
                fdata['path_segments'].insert(0, root_path)

        root_data = self._create_empty_resource_data()

        if self.has_single_resource:
            # single resource with one or more segments
            rdata = files_data[0]
            root_path = rdata['path']
            rdata = remove_properties_and_basics(rdata)
            root_data.update(rdata)

        # Create root resource
        root = self._create_root_resource(
            name=file_name(root_path),
            path=root_path,
            is_file=root_is_file,
        )

        for name, value in root_data.items():
            # skip known properties
            if name not in KNOW_PROPS:
                setattr(root, name, value)

        if TRACE:
            logger_debug('VirtualCodebase.populate: root:', root)

        # TODO: report error if filtering the root with a paths?
        self.save_resource(root)

        if self.has_single_resource:
            if TRACE:
                logger_debug('VirtualCodebase.populate: with single resource.')
            return

        all_paths = None
        if self.paths:
            # build a set of all all paths and all their ancestors
            all_paths = set()
            for path in self.paths:
                all_paths.update(get_ancestor_paths(path, include_self=True))

        # Create other Resources from scan info

        # Note that we do not know the ordering there.
        # Therefore we sort in place by path segments
        files_data.sort(key=itemgetter('path_segments'))

        # We create directories that exist in the scan or create these that
        # exist only in paths
        duplicated_paths = set()
        last_path = None
        for fdata in files_data:
            path = fdata.get('path')

            # skip the ones we did not request
            if all_paths and path not in all_paths:
                continue

            # these are no longer needed
            path_segments = fdata.pop('path_segments')

            if not last_path:
                last_path = path
            elif last_path == path:
                duplicated_paths.add(path)
            else:
                last_path = path

            name = fdata.get('name', None) or None
            if not name:
                name = file_name(path)

            is_file = fdata.get('type', 'file') == 'file'

            parent = self._get_parent_directory(path_segments=path_segments)
            resource = self._get_or_create_resource(
                name=name,
                path=path,
                parent=parent,
                is_file=is_file,
            )
            # set data
            for name, value in fdata.items():
                # skip known properties
                if name not in KNOW_PROPS:
                    setattr(resource, name, value)

            self.save_resource(resource)

        if duplicated_paths:
            raise Exception(
                'Illegal combination of VirtualCode multiple inputs: '
                f'duplicated paths: {list(duplicated_paths)}',
            )

    def _get_parent_directory(self, path_segments):
        """
        Ensure that all directories in a sequence of path_segments exist
        and return the last one.
        """
        # TODO: handle single resource codebases
        resources_by_path = self.resources_by_path

        # remove the first which is the root, already created
        # and the last which is the current "child" segment
        path_segments = path_segments[1:-1]

        current = self.root
        for segment in path_segments:
            existing = resources_by_path.get(segment)
            if not existing:
                existing = self._get_or_create_resource(
                    name=segment,
                    # build the path based on parent
                    path=posixpath_join(current.path, segment),
                    parent=current,
                    is_file=False,
                )
            current = existing
        return current

    def _create_root_resource(self, name, path, is_file):
        """
        Create and return the root Resource of this codebase.
        """
        # we cannot recreate a root if it exists!!
        if self.root:
            raise TypeError('Root resource already exists and cannot be recreated')

        path = clean_path(path)

        if TRACE:
            logger_debug(f'  VirtualCodebase._create_root_resource: {path!r} is_file: {is_file}')

        root = self.resource_class(
            name=name,
            location=None,
            # never cached
            cache_location=None,
            path=path,
            is_root=True,
            is_file=is_file,
        )

        self.resources_by_path[path] = root
        self.resources_count += 1
        self.root = root
        return root


KNOW_PROPS = set(['type', 'base_name', 'extension', 'path', 'name', 'path_segments'])


def remove_properties_and_basics(resource_data):
    """
    Given a mapping of resource_data attributes to use as "kwargs", return a new
    mapping with the known properties removed.
    """
    return {k: v for k, v in resource_data.items() if k not in KNOW_PROPS}


def get_ancestor_paths(path, include_self=False):
    """
    Yield all subpaths from a POSIX path.

    For example::
    >>> path = 'foo/bar/baz'
    >>> results = list(get_ancestor_paths(path))
    >>> assert results == ['foo', 'foo/bar'], results
    >>> results = list(get_ancestor_paths(path, include_self=True))
    >>> assert results == ['foo', 'foo/bar', 'foo/bar/baz'], results
    >>> results = list(get_ancestor_paths('foo', include_self=False))
    >>> assert results == [], results
    """
    assert path
    segments = path.split('/')
    if not include_self:
        segments = segments[:-1]
    subpath = []
    for segment in segments:
        subpath.append(segment)
        yield '/'.join(subpath)



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from unittest import TestCase

from commoncode.codec import bin_to_num
from commoncode.codec import num_to_bin
from commoncode.codec import urlsafe_b64encode_int


class TestCodec(TestCase):

    def test_bin_to_num_basic(self):
        expected = 123
        result = bin_to_num(b'{')
        assert result == expected

    def test_bin_to_num_zero(self):
        expected = 0
        result = bin_to_num(b'\x00')
        assert result == expected

    def test_bin_to_num_large_number(self):
        expected = 432346237462348763
        result = bin_to_num(b'\x06\x00\x00\x9c\xbf\xeb\x83\xdb')
        assert result == expected

    def test_bin_to_num_and_num_to_bin_is_idempotent(self):
        expected = 432346237462348763
        result = bin_to_num(num_to_bin(432346237462348763))
        assert result == expected

    def test_num_to_bin_basic(self):
        expected = b'{'
        result = num_to_bin(123)
        assert result == expected

    def test_num_to_bin_zero(self):
        expected = b'\x00'
        result = num_to_bin(0)
        assert result == expected

    def test_num_to_bin_large_number(self):
        expected = b'\x06\x00\x00\x9c\xbf\xeb\x83\xdb'
        result = num_to_bin(432346237462348763)
        assert result == expected

    def test_num_to_bin_bin_to_num_is_idempotent(self):
        expected = b'\x06\x00\x00\x9c\xbf\xeb\x83\xdb'
        result = num_to_bin(bin_to_num(b'\x06\x00\x00\x9c\xbf\xeb\x83\xdb'))
        assert result == expected

    def test_urlsafe_b64encode_int_zero(self):
        assert urlsafe_b64encode_int(0) ==  b'AA=='

    def test_urlsafe_b64encode_int_basic(self):
        assert urlsafe_b64encode_int(123123123123) == b'HKq1w7M='

    def test_urlsafe_b64encode_int_limit_8bits_255(self):
        assert urlsafe_b64encode_int(255) == b'_w=='

    def test_urlsafe_b64encode_int_limit_8bits_256(self):
        assert urlsafe_b64encode_int(256) == b'AQA='

    def test_urlsafe_b64encode_int_adds_no_padding_for_number_that_are_multiple_of_6_bits(self):
        assert urlsafe_b64encode_int(0xFFFFFFFFFFFFFFFFFF) == b'____________'
        assert len(urlsafe_b64encode_int(0xFFFFFFFFFFFF)) == 8

    def test_urlsafe_b64encode_int_very_large_number(self):
        b64 = (b'QAAAAAAgAAAAAQAACAAAAAAAAAAAAAAkAAIAAAAAAAAAAAAAAACAAIAAAAAAAAAAAAAAAAAAAAAAAAAAAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAIAAAAAAAAAAAAAAAAAAAAiAAAAAAAIAAAAAAAAAAAAAAEAACAAAAAAAA=')
        expected = b64
        num = 2678771517966886466622496485850735537232223496190189203248435106535830319026141316924949516664780383591425235756710588949364368366679435700855700642969357960349427980681242720502045830438444033569999428606714388704082526548154984676817460705606960919023941301616034362869262429593297635158449513824256
        result = urlsafe_b64encode_int(num)
        assert result == expected



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from collections import Counter
from unittest.case import TestCase

from commoncode.functional import flatten
from commoncode.functional import memoize
from commoncode.functional import partial


class TestFunctional(TestCase):

    def test_flatten(self):
        expected = [7, 6, 5, 4, 'a', 3, 3, 2, 1]
        test = flatten([7, (6, [5, [4, ["a"], 3]], 3), 2, 1])
        assert test == expected

    def test_flatten_generator(self):

        def gen():
            for _ in range(2):
                yield range(5)

        expected = [0, 1, 2, 3, 4, 0, 1, 2, 3, 4]
        test = flatten(gen())
        assert test == expected

    def test_flatten_empties(self):
        expected = ['a']
        test = flatten([[], (), ['a']])
        assert test == expected

    def test_partial(self):

        def test_func(a, b):
            pass

        wrapped = partial(test_func, a=2)
        assert wrapped.__name__ == 'test_func'

    def test_memoized(self):
        call_count = Counter()

        @memoize
        def test_func(a):
            call_count[a] += 1

        test_func(1)
        assert 1 == call_count[1]
        test_func(1)
        assert 1 == call_count[1]
        test_func(2)
        assert 1 == call_count[2]
        test_func(2)
        assert 1 == call_count[2]
        test_func(2)
        assert 1 == call_count[1]



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import attr
from attr.validators import in_ as choices  # NOQA
import typing

"""
Utilities and helpers for data classes.
"""

HELP_METADATA = '__field_help'
LABEL_METADATA = '__field_label'


def attribute(default=attr.NOTHING, validator=None,
              repr=False, eq=True, order=True,  # NOQA
              init=True, type=None, converter=None,  # NOQA
              help=None, label=None, metadata=None,):  # NOQA
    """
    A generic attribute with help metadata and that is not included in the
    representation by default.
    """
    metadata = metadata or dict()
    if help:
        metadata[HELP_METADATA] = help

    if label:
        metadata[LABEL_METADATA] = label

    return attr.attrib(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=init,
        metadata=metadata,
        type=type,
        converter=converter
    )


def Boolean(default=False, validator=None, repr=False, eq=True, order=True,  # NOQA
            converter=None, label=None, help=None,):  # NOQA
    """
    A boolean attribute.
    """
    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=bool,
        converter=converter,
        help=help,
        label=label,
    )


def TriBoolean(default=None, validator=None, repr=False, eq=True, order=True,  # NOQA
            converter=None, label=None, help=None,):  # NOQA
    """
    A tri-boolean attribute with possible values of None, True and False.
    """
    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=bool,
        converter=converter,
        help=help,
        label=label,
    )


def String(default=None, validator=None, repr=False, eq=True, order=True,  # NOQA
           converter=None, label=None, help=None,):  # NOQA
    """
    A string attribute.
    """
    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=str,
        converter=converter,
        help=help,
        label=label,
    )


def Integer(default=0, validator=None, repr=False, eq=True, order=True,  # NOQA
            converter=None, label=None, help=None,):  # NOQA
    """
    An integer attribute.
    """
    converter = converter or attr.converters.optional(int)
    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=int,
        converter=converter,
        help=help,
        label=label,
    )


def Float(default=0.0, validator=None, repr=False, eq=True, order=True,  # NOQA
          converter=None, label=None, help=None,):  # NOQA
    """
    A float attribute.
    """
    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=float,
        converter=converter,
        help=help,
        label=label,
    )


def List(item_type=typing.Any, default=attr.NOTHING, validator=None,
         repr=False, eq=True, order=True,  # NOQA
         converter=None, label=None, help=None,):  # NOQA
    """
    A list attribute: the optional item_type defines the type of items it stores.
    """
    if default is attr.NOTHING:
        default = attr.Factory(list)

    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=typing.List[item_type],
        converter=converter,
        help=help,
        label=label,
    )


def Mapping(value_type=typing.Any, default=attr.NOTHING, validator=None,
            repr=False, eq=True, order=True,  # NOQA
            converter=None, help=None, label=None):  # NOQA
    """
    A mapping attribute: the optional value_type defines the type of values it
    stores. The key is always a string.

    Notes: in Python 2 the type is Dict as there is no typing available for
    dict for now.
    """
    if default is attr.NOTHING:
        default = attr.Factory(dict)

    return attribute(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        init=True,
        type=typing.Dict[str, value_type],
        converter=converter,
        help=help,
        label=label,
    )

##################################################
# FIXME: add proper support for dates!!!
##################################################


def Date(default=None, validator=None, repr=False, eq=True, order=True,  # NOQA
           converter=None, label=None, help=None,):  # NOQA
    """
    A date attribute. It always serializes to an ISO date string.
    Behavior is TBD and for now this is exactly a string.
    """
    return String(
        default=default,
        validator=validator,
        repr=repr,
        eq=eq,
        order=order,
        converter=converter,
        help=help,
        label=label,
    )



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import binascii
import hashlib

from commoncode.codec import bin_to_num
from commoncode.codec import urlsafe_b64encode
from commoncode import filetype

"""
Hashes and checksums.

Low level hash functions using standard crypto hashes used to construct hashes
of various lengths. Hashes that are smaller than 128 bits are based on a
truncated md5. Other length use SHA hashes.

Checksums are operating on files.
"""


def _hash_mod(bitsize, hmodule):
    """
    Return a hashing class returning hashes with a `bitsize` bit length. The
    interface of this class is similar to the hash module API.
    """

    class hasher(object):

        def __init__(self, msg=None):
            self.digest_size = bitsize // 8
            self.h = msg and hmodule(msg).digest()[:self.digest_size] or None

        def digest(self):
            return bytes(self.h)

        def hexdigest(self):
            return self.h and binascii.hexlify(self.h).decode('utf-8')

        def b64digest(self):
            return self.h and urlsafe_b64encode(self.h).decode('utf-8')

        def intdigest(self):
            return self.h and int(bin_to_num(self.h))

    return hasher


# Base hashers for each bit size
_hashmodules_by_bitsize = {
    # md5-based
    32: _hash_mod(32, hashlib.md5),
    64: _hash_mod(64, hashlib.md5),
    128: _hash_mod(128, hashlib.md5),
    # sha-based
    160: _hash_mod(160, hashlib.sha1),
    256: _hash_mod(256, hashlib.sha256),
    384: _hash_mod(384, hashlib.sha384),
    512: _hash_mod(512, hashlib.sha512)
}


def get_hasher(bitsize):
    """
    Return a hasher for a given size in bits of the resulting hash.
    """
    return _hashmodules_by_bitsize[bitsize]


class sha1_git_hasher(object):
    """
    Hash content using the git blob SHA1 convention.
    """

    def __init__(self, msg=None):
        self.digest_size = 160 // 8
        self.h = msg and self._compute(msg) or None

    def _compute(self, msg):
        # note: bytes interpolation is new in Python 3.5
        git_blob_msg = b'blob %d\0%s' % (len(msg), msg)
        return hashlib.sha1(git_blob_msg).digest()

    def digest(self):
        return bytes(self.h)

    def hexdigest(self):
        return self.h and binascii.hexlify(self.h).decode('utf-8')

    def b64digest(self):
        return self.h and urlsafe_b64encode(self.h).decode('utf-8')

    def intdigest(self):
        return self.h and int(bin_to_num(self.h))


_hashmodules_by_name = {
    'md5': get_hasher(128),
    'sha1': get_hasher(160),
    'sha1_git': sha1_git_hasher,
    'sha256': get_hasher(256),
    'sha384': get_hasher(384),
    'sha512': get_hasher(512)
}


def checksum(location, name, base64=False):
    """
    Return a checksum of `bitsize` length from the content of the file at
    `location`. The checksum is a hexdigest or base64-encoded is `base64` is
    True.
    """
    if not filetype.is_file(location):
        return
    hasher = _hashmodules_by_name[name]

    # fixme: we should read in chunks?
    with open(location, 'rb') as f:
        hashable = f.read()

    hashed = hasher(hashable)
    if base64:
        return hashed.b64digest()

    return hashed.hexdigest()


def md5(location):
    return checksum(location, name='md5', base64=False)


def sha1(location):
    return checksum(location, name='sha1', base64=False)


def b64sha1(location):
    return checksum(location, name='sha1', base64=True)


def sha256(location):
    return checksum(location, name='sha256', base64=False)


def sha512(location):
    return checksum(location, name='sha512', base64=False)


def sha1_git(location):
    return checksum(location, name='sha1_git', base64=False)


def multi_checksums(location, checksum_names=('md5', 'sha1', 'sha256', 'sha512', 'sha1_git')):
    """
    Return a mapping of hexdigest checksums keyed by checksum name from the content
    of the file at `location`. Use the `checksum_names` list of checksum names.
    The mapping is guaranted to contains all the requested names as keys.
    If the location is not a file, the values are None.
    """
    results = dict([(name, None) for name in checksum_names])
    if not filetype.is_file(location):
        return results

    # fixme: we should read in chunks?
    with open(location, 'rb') as f:
        hashable = f.read()

    for name in checksum_names:
        results[name] = _hashmodules_by_name[name](hashable).hexdigest()
    return results



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os

from commoncode.testcase import FileDrivenTesting
from commoncode.cliutils import progressmanager

class TestProgressBar(FileDrivenTesting):

    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def test_progressmanager_works(self):
        iterable = range(10)
        with progressmanager(iterable) as it:
            for _ in it:
                pass

    def test_progressmanager_verbose(self):
        iterable = range(10)
        with progressmanager(iterable, verbose=True) as it:
            for _ in it:
                pass



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from datetime import datetime

from commoncode.testcase import FileBasedTesting
from commoncode.timeutils import time2tstamp
from commoncode.timeutils import tstamp2time
from commoncode.timeutils import UTC


class TestTimeStamp(FileBasedTesting):

    def test_time2tstamp_is_path_safe_and_file_is_writable(self):
        ts = time2tstamp()
        tf = self.get_temp_file(extension='ext', dir_name=ts, file_name=ts)
        fd = open(tf, 'w')
        fd.write('a')
        fd.close()

    def test_time2tstamp_accepts_existing_datetimes(self):
        ts = time2tstamp()
        tf = self.get_temp_file(extension='ext', dir_name=ts, file_name=ts)
        fd = open(tf, 'w')
        fd.write('a')
        fd.close()

    def test_time2tstamp_raises_on_non_datetime(self):
        self.assertRaises(AttributeError, time2tstamp, 'some')
        self.assertRaises(AttributeError, time2tstamp, 1)

    def test_time2tstamp_tstamp2time_is_idempotent(self):
        dt = datetime.utcnow()
        ts = time2tstamp(dt)
        dt_from_ts = tstamp2time(ts)
        assert dt_from_ts == dt

    def test_tstamp2time_format(self):
        import re
        ts = time2tstamp()
        pat = r'^20\d\d-[0-1][0-9]-[0-3]\dT[0-2]\d[0-6]\d[0-6]\d.\d\d\d\d\d\d$'
        assert re.match(pat, ts)

    def test_tstamp2time(self):
        dt_from_ts = tstamp2time('2010-11-12T131415.000016')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=16, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time2(self):
        dt_from_ts = tstamp2time('20101112T131415.000016')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=16, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time3(self):
        dt_from_ts = tstamp2time('20101112T131415.000016Z')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=16, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time4(self):
        dt_from_ts = tstamp2time('2010-11-12T131415')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=0, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time5(self):
        dt_from_ts = tstamp2time('2010-11-12T13:14:15')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=0, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time6(self):
        dt_from_ts = tstamp2time('20101112T13:14:15')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=0, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time7(self):
        dt_from_ts = tstamp2time('20101112T13:14:15Z')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=0, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time8(self):
        dt_from_ts = tstamp2time('20101112T13:14:15Z')
        assert datetime(year=2010, month=11, day=12, hour=13, minute=14, second=15, microsecond=0, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time9(self):
        dt_from_ts = tstamp2time('2010-06-30T21:26:40.000Z')
        assert datetime(year=2010, month=6, day=30, hour=21, minute=26, second=40, microsecond=0, tzinfo=UTC()) == dt_from_ts

    def test_tstamp2time_raise(self):
        self.assertRaises(ValueError, tstamp2time, '201011A12T13:14:15Z')



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import ntpath
import posixpath
import re

from os.path import commonprefix

from commoncode.text import as_unicode
from commoncode.text import toascii
from commoncode.fileutils import as_posixpath
from commoncode.fileutils import as_winpath
from commoncode.fileutils import is_posixpath

"""
Various path utilities such as common prefix and suffix functions, conversion
to OS-safe paths and to POSIX paths.
"""
#
# Build OS-portable and safer paths


def safe_path(path, posix=False, preserve_spaces=False):
    """
    Convert `path` to a safe and portable POSIX path usable on multiple OSes.
    The returned path is an ASCII-only byte string, resolved for relative
    segments and itself relative.

    The `path` is treated as a POSIX path if `posix` is True or as a Windows
    path with blackslash separators otherwise.

    If `preserve_spaces` is True, then the spaces in `path` will not be replaced.
    """
    # if the path is UTF, try to use unicode instead
    if not isinstance(path, str):
        path = as_unicode(path)

    path = path.strip()

    if not is_posixpath(path):
        path = as_winpath(path)
        posix = False

    path = resolve(path, posix)

    _pathmod, path_sep = path_handlers(path, posix)

    segments = [s.strip() for s in path.split(path_sep) if s.strip()]
    segments = [portable_filename(s, preserve_spaces=preserve_spaces) for s in segments]

    if not segments:
        return '_'

    # always return posix
    path = '/'.join(segments)
    return as_posixpath(path)


def path_handlers(path, posix=True):
    """
    Return a path module and path separator to use for handling (e.g. split and
    join) `path` using either POSIX or Windows conventions depending on the
    `path` content. Force usage of POSIX conventions if `posix` is True.
    """
    # determine if we use posix or windows path handling
    is_posix = is_posixpath(path)
    use_posix = posix or is_posix
    pathmod = use_posix and posixpath or ntpath
    path_sep = '/' if use_posix else '\\'
    return pathmod, path_sep


def resolve(path, posix=True):
    """
    Return a resolved relative POSIX path from `path` where extra slashes
    including leading and trailing slashes are removed, dot '.' and dotdot '..'
    path segments have been removed or resolved as possible. When a dotdot path
    segment cannot be further resolved and would be "escaping" from the provided
    path "tree", it is replaced by the string 'dotdot'.

    The `path` is treated as a POSIX path if `posix` is True (default) or as a
    Windows path with blackslash separators otherwise.
    """
    if not path:
        return '.'

    path = path.strip()
    if not path:
        return '.'

    if not is_posixpath(path):
        path = as_winpath(path)
        posix = False

    pathmod, path_sep = path_handlers(path, posix)

    path = path.strip(path_sep)
    segments = [s.strip() for s in path.split(path_sep) if s.strip()]

    # remove empty (// or ///) or blank (space only) or single dot segments
    segments = [s for s in segments if s and s != '.']

    path = path_sep.join(segments)

    # resolves . dot, .. dotdot
    path = pathmod.normpath(path)

    segments = path.split(path_sep)

    # remove empty or blank segments
    segments = [s.strip() for s in segments if s and s.strip()]

    # is this a windows absolute path? if yes strip the colon to make this relative
    if segments and len(segments[0]) == 2 and segments[0].endswith(':'):
        segments[0] = segments[0][:-1]

    # replace any remaining (usually leading) .. segment with a literal "dotdot"
    dotdot = 'dotdot'
    dd = '..'
    segments = [dotdot if s == dd else s for s in segments if s]
    if segments:
        path = path_sep.join(segments)
    else:
        path = '.'

    path = as_posixpath(path)

    return path


legal_punctuation = r"!\#$%&\(\)\+,\-\.;\=@\[\]_\{\}\~"
legal_spaces = r" "
legal_chars = r'A-Za-z0-9' + legal_punctuation
legal_chars_inc_spaces = legal_chars + legal_spaces
illegal_chars_re = r'[^' + legal_chars + r']'
illegal_chars_exc_spaces_re = r'[^' + legal_chars_inc_spaces + r']'
replace_illegal_chars = re.compile(illegal_chars_re).sub
replace_illegal_chars_exc_spaces = re.compile(illegal_chars_exc_spaces_re).sub


def portable_filename(filename, preserve_spaces=False):
    """
    Return a new name for `filename` that is portable across operating systems.

    In particular the returned file name is guaranteed to be:
    - a portable name on most OSses using a limited ASCII characters set including
      some limited punctuation.
    - a valid name on Linux, Windows and Mac.

    Unicode file names are transliterated to plain ASCII.

    See for more details:
    - http://www.opengroup.org/onlinepubs/007904975/basedefs/xbd_chap03.html
    - https://msdn.microsoft.com/en-us/library/windows/desktop/aa365247(v=vs.85).aspx
    - http://www.boost.org/doc/libs/1_36_0/libs/filesystem/doc/portability_guide.htm

    Also inspired by Werkzeug:
    https://raw.githubusercontent.com/pallets/werkzeug/8c2d63ce247ba1345e1b9332a68ceff93b2c07ab/werkzeug/utils.py

    If `preserve_spaces` is True, then spaces in `filename` will not be replaced.
    """
    filename = toascii(filename, translit=True)

    if not filename:
        return '_'

    if preserve_spaces:
        filename = replace_illegal_chars_exc_spaces('_', filename)
    else:
        filename = replace_illegal_chars('_', filename)

    # these are illegal both upper and lowercase and with or without an extension
    # we insert an underscore after the base name.
    windows_illegal_names = set([
        'com1', 'com2', 'com3', 'com4', 'com5', 'com6', 'com7', 'com8', 'com9',
        'lpt1', 'lpt2', 'lpt3', 'lpt4', 'lpt5', 'lpt6', 'lpt7', 'lpt8', 'lpt9',
        'aux', 'con', 'nul', 'prn'
    ])

    basename, dot, extension = filename.partition('.')
    if basename.lower() in windows_illegal_names:
        filename = ''.join([basename, '_', dot, extension])

    # no name made only of dots.
    if set(filename) == set(['.']):
        filename = 'dot' * len(filename)

    # replaced any leading dotdot
    if filename != '..' and filename.startswith('..'):
        while filename.startswith('..'):
            filename = filename.replace('..', '__', 1)

    return filename

#
# paths comparisons, common prefix and suffix extraction
#


def common_prefix(s1, s2):
    """
    Return the common leading subsequence of two sequences and its length.
    """
    if not s1 or not s2:
        return None, 0
    common = commonprefix((s1, s2,))
    if common:
        return common, len(common)
    else:
        return None, 0


def common_suffix(s1, s2):
    """
    Return the common trailing subsequence between two sequences and its length.
    """
    if not s1 or not s2:
        return None, 0
    # revert the seqs and get a common prefix
    common, lgth = common_prefix(s1[::-1], s2[::-1])
    # revert again
    common = common[::-1] if common else common
    return common, lgth


def common_path_prefix(p1, p2):
    """
    Return the common leading path between two posix paths and the number of
    matched path segments.
    """
    return _common_path(p1, p2, common_func=common_prefix)


def common_path_suffix(p1, p2):
    """
    Return the common trailing path between two posix paths and the number of
    matched path segments.
    """
    return _common_path(p1, p2, common_func=common_suffix)


def split(p):
    """
    Split a posix path in a sequence of segments, ignoring leading and trailing
    slash. Return an empty sequence for an empty path and the root path /.
    """
    if not p:
        return []
    p = p.strip('/').split('/')
    return [] if p == [''] else p


def _common_path(p1, p2, common_func):
    """
    Return a common leading or trailing path brtween paths `p1` and `p2` and the
    common length in number of segments using the `common_func` path comparison
    function.
    """
    common, lgth = common_func(split(p1), split(p2))
    common = '/'.join(common) if common else None
    return common, lgth



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import filecmp
import json
import os
import shutil
import stat
import sys
from os import path
from collections import defaultdict
from itertools import chain
from unittest import TestCase as TestCaseClass

import saneyaml

from commoncode import fileutils
from commoncode import filetype
from commoncode.archive import extract_tar
from commoncode.archive import extract_tar_raw
from commoncode.archive import extract_tar_uni
from commoncode.archive import extract_zip
from commoncode.archive import extract_zip_raw
from commoncode.archive import tar_can_extract  # NOQA
from commoncode.system import on_posix
from commoncode.system import on_windows

# a base test dir specific to a given test run
# to ensure that multiple tests run can be launched in parallel
test_run_temp_dir = None

# set to 1 to see the slow tests
timing_threshold = sys.maxsize


def to_os_native_path(path):
    """
    Normalize a path to use the native OS path separator.
    """
    OS_PATH_SEP = '\\' if on_windows else '/'

    return (
        path.replace('/', OS_PATH_SEP)
        .replace(u'\\', OS_PATH_SEP)
        .rstrip(OS_PATH_SEP)
    )


def get_test_loc(
    test_path,
    test_data_dir,
    debug=False,
    must_exist=True,
):
    """
    Given a `test_path` relative to the `test_data_dir` directory, return the
    location to a test file or directory for this path. No copy is done.
    Raise an IOError if `must_exist` is True and the `test_path` does not exists.
    """
    if debug:
        import inspect
        caller = inspect.stack()[1][3]
        print('\nget_test_loc,%(caller)s,"%(test_path)s","%(test_data_dir)s"' % locals())

    assert test_path
    assert test_data_dir

    if not path.exists(test_data_dir):
        raise IOError("[Errno 2] No such directory: test_data_dir not found:"
                      " '%(test_data_dir)s'" % locals())

    tpath = to_os_native_path(test_path)
    test_loc = path.abspath(path.join(test_data_dir, tpath))

    if must_exist and not path.exists(test_loc):
        raise IOError("[Errno 2] No such file or directory: "
                      "test_path not found: '%(test_loc)s'" % locals())

    return test_loc


class FileDrivenTesting(object):
    """
    Add support for handling test files and directories, including managing
    temporary test resources and doing file-based assertions.
    This can be used as a standalone object if needed.
    """
    test_data_dir = None

    def get_test_loc(self, test_path, copy=False, debug=False, must_exist=True):
        """
        Given a `test_path` relative to the self.test_data_dir directory, return the
        location to a test file or directory for this path. Copy to a temp
        test location if `copy` is True.

        Raise an IOError if `must_exist` is True and the `test_path` does not
        exists.
        """
        test_data_dir = self.test_data_dir
        if debug:
            import inspect
            caller = inspect.stack()[1][3]
            print('\nself.get_test_loc,%(caller)s,"%(test_path)s"' % locals())

        test_loc = get_test_loc(
            test_path,
            test_data_dir,
            debug=debug,
            must_exist=must_exist,
        )
        if copy:
            base_name = path.basename(test_loc)
            if filetype.is_file(test_loc):
                # target must be an existing dir
                target_dir = self.get_temp_dir()
                fileutils.copyfile(test_loc, target_dir)
                test_loc = path.join(target_dir, base_name)
            else:
                # target must be a NON existing dir
                target_dir = path.join(self.get_temp_dir(), base_name)
                fileutils.copytree(test_loc, target_dir)
                # cleanup of VCS that could be left over from checkouts
                self.remove_vcs(target_dir)
                test_loc = target_dir
        return test_loc

    def get_temp_file(self, extension=None, dir_name='td', file_name='tf'):
        """
        Return a unique new temporary file location to a non-existing temporary
        file that can safely be created without a risk of name collision.
        """
        if extension is None:
            extension = '.txt'

        if extension and not extension.startswith('.'):
                extension = '.' + extension

        file_name = file_name + extension
        temp_dir = self.get_temp_dir(dir_name)
        location = path.join(temp_dir, file_name)
        return location

    def get_temp_dir(self, sub_dir_path=None):
        """
        Create a unique new temporary directory location. Create directories
        identified by sub_dir_path if provided in this temporary directory.
        Return the location for this unique directory joined with the
        `sub_dir_path` if any.
        """
        # ensure that we have a new unique temp directory for each test run
        global test_run_temp_dir
        if not test_run_temp_dir:
            import tempfile
            test_tmp_root_dir = tempfile.gettempdir()
            # now we add a space in the path for testing path with spaces
            test_run_temp_dir = fileutils.get_temp_dir(
                base_dir=test_tmp_root_dir, prefix='scancode-tk-tests -')

        test_run_temp_subdir = fileutils.get_temp_dir(
            base_dir=test_run_temp_dir, prefix='')

        if sub_dir_path:
            # create a sub directory hierarchy if requested
            sub_dir_path = to_os_native_path(sub_dir_path)
            test_run_temp_subdir = path.join(test_run_temp_subdir, sub_dir_path)
            fileutils.create_dir(test_run_temp_subdir)
        return test_run_temp_subdir

    def remove_vcs(self, test_dir):
        """
        Remove some version control directories and some temp editor files.
        """
        vcses = ('CVS', '.svn', '.git', '.hg')
        for root, dirs, files in os.walk(test_dir):
            for vcs_dir in vcses:
                if vcs_dir in dirs:
                    for vcsroot, vcsdirs, vcsfiles in os.walk(test_dir):
                        for vcsfile in vcsdirs + vcsfiles:
                            vfile = path.join(vcsroot, vcsfile)
                            fileutils.chmod(vfile, fileutils.RW, recurse=False)
                    shutil.rmtree(path.join(root, vcs_dir), False)

            # editors temp file leftovers
            tilde_files = [path.join(root, file_loc)
                           for file_loc in files if file_loc.endswith('~')]
            for tf in tilde_files:
                os.remove(tf)

    def __extract(self, test_path, extract_func=None, verbatim=False):
        """
        Given an archive file identified by test_path relative
        to a test files directory, return a new temp directory where the
        archive file has been extracted using extract_func.
        If `verbatim` is True preserve the permissions.
        """
        assert test_path and test_path != ''
        test_path = to_os_native_path(test_path)
        target_path = path.basename(test_path)
        target_dir = self.get_temp_dir(target_path)
        original_archive = self.get_test_loc(test_path)
        extract_func(original_archive, target_dir, verbatim=verbatim)
        return target_dir

    def extract_test_zip(self, test_path, *args, **kwargs):
        return self.__extract(test_path, extract_zip)

    def extract_test_zip_raw(self, test_path, *args, **kwargs):
        return self.__extract(test_path, extract_zip_raw)

    def extract_test_tar(self, test_path, verbatim=False):
        return self.__extract(test_path, extract_tar, verbatim)

    def extract_test_tar_raw(self, test_path, *args, **kwargs):
        return self.__extract(test_path, extract_tar_raw)

    def extract_test_tar_unicode(self, test_path, *args, **kwargs):
        return self.__extract(test_path, extract_tar_uni)


class FileBasedTesting(TestCaseClass, FileDrivenTesting):
    pass


class dircmp(filecmp.dircmp):
    """
    Compare the content of dir1 and dir2. In contrast with filecmp.dircmp,
    this subclass also compares the content of files with the same path.
    """

    def phase3(self):
        """
        Find out differences between common files.
        Ensure we are using content comparison, not os.stat-only.
        """
        comp = filecmp.cmpfiles(self.left, self.right, self.common_files, shallow=False)
        self.same_files, self.diff_files, self.funny_files = comp


def is_same(dir1, dir2):
    """
    Compare two directory trees for structure and file content.
    Return False if they differ, True is they are the same.
    """
    compared = dircmp(dir1, dir2)
    if (compared.left_only or compared.right_only or compared.diff_files
        or compared.funny_files):
        return False

    for subdir in compared.common_dirs:
        if not is_same(path.join(dir1, subdir),
                       path.join(dir2, subdir)):
            return False
    return True


def file_cmp(file1, file2, ignore_line_endings=False):
    """
    Compare two files content.
    Return False if they differ, True is they are the same.
    """
    with open(file1, 'rb') as f1:
        f1c = f1.read()
        if ignore_line_endings:
            f1c = b'\n'.join(f1c.splitlines(False))
    with open(file2, 'rb') as f2:
        f2c = f2.read()
        if ignore_line_endings:
            f2c = b'\n'.join(f2c.splitlines(False))
    assert f2c == f1c


def make_non_readable(location):
    """
    Make location non readable for tests purpose.
    """
    if on_posix:
        current_stat = stat.S_IMODE(os.lstat(location).st_mode)
        os.chmod(location, current_stat & ~stat.S_IREAD)
    else:
        os.chmod(location, 0o555)


def make_non_writable(location):
    """
    Make location non writable for tests purpose.
    """
    if on_posix:
        current_stat = stat.S_IMODE(os.lstat(location).st_mode)
        os.chmod(location, current_stat & ~stat.S_IWRITE)
    else:
        make_non_readable(location)


def make_non_executable(location):
    """
    Make location non executable for tests purpose.
    """
    if on_posix:
        current_stat = stat.S_IMODE(os.lstat(location).st_mode)
        os.chmod(location, current_stat & ~stat.S_IEXEC)


def get_test_file_pairs(test_dir):
    """
    Yield tuples of (data_file, test_file) from a test data `test_dir` directory.
    Raise exception for orphaned/dangling files.
    Each test consist of a pair of files:
    - a test file.
    - a data file with the same name as a test file and a '.yml' extension added.
    Each test file path should be unique in the tree ignoring case.
    """
    # collect files with .yml extension and files with other extensions
    data_files = {}
    test_files = {}
    dangling_test_files = set()
    dangling_data_files = set()
    paths_ignoring_case = defaultdict(list)

    for top, _, files in os.walk(test_dir):
        for tfile in files:
            if tfile.endswith('~'):
                continue
            file_path = path.abspath(path.join(top, tfile))

            if tfile.endswith('.yml'):
                data_file_path = file_path
                test_file_path = file_path.replace('.yml', '')
            else:
                test_file_path = file_path
                data_file_path = test_file_path + '.yml'

            if not path.exists(test_file_path):
                dangling_test_files.add(test_file_path)

            if not path.exists(data_file_path):
                dangling_data_files.add(data_file_path)

            paths_ignoring_case[file_path.lower()].append(file_path)

            data_files[test_file_path] = data_file_path
            test_files[test_file_path] = test_file_path

    # ensure that we haev no dangling files
    if dangling_test_files or dangling_data_files:
        msg = ['Dangling missing test files without a YAML data file:'] + sorted(dangling_test_files)
        msg += ['Dangling missing YAML data files without a test file'] + sorted(dangling_data_files)
        msg = '\n'.join(msg)
        print(msg)
        raise Exception(msg)

    # ensure that each data file has a corresponding test file
    diff = set(data_files.keys()).symmetric_difference(set(test_files.keys()))
    if diff:
        msg = [
            'Orphaned copyright test file(s) found: '
            'test file without its YAML test data file '
            'or YAML test data file without its test file.'] + sorted(diff)
        msg = '\n'.join(msg)
        print(msg)
        raise Exception(msg)

    # ensure that test file paths are unique when you ignore case
    # we use the file names as test method names (and we have Windows that's
    # case insensitive
    dupes = list(chain.from_iterable(
        paths for paths in paths_ignoring_case.values() if len(paths) != 1))
    if dupes:
        msg = ['Non unique test/data file(s) found when ignoring case!'] + sorted(dupes)

        msg = '\n'.join(msg)
        print(msg)
        raise Exception(msg)

    for test_file in test_files:
        yield test_file + '.yml', test_file


def check_against_expected_json_file(results, expected_file, regen=False):
    """
    Check that the ``results`` data are the same as the data in the
    ``expected_file`` expected JSON data file.

    If `regen` is True the expected_file will overwritten with the ``results``.
    This is convenient for updating tests expectations. But use with caution.
    """
    if regen:
        with open(expected_file, 'w') as reg:
            json.dump(results, reg, indent=2, separators=(',', ': '))
        expected = results
    else:
        with open(expected_file) as exp:
            expected = json.load(exp)

    # NOTE we redump the JSON as a YAML string for easier display of
    # the failures comparison/diff
    if results != expected:
        expected = saneyaml.dump(expected)
        results = saneyaml.dump(results)
        assert results == expected



# -*- coding: utf-8 -*-
#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from commoncode import text
from commoncode.text import CR
from commoncode.text import LF


def test_lines():
    t = '''This problem is.
It is therefore


 However,we
without introducing ..
 However, I have


'''
    assert len([p[1] for p in text.lines(t)]) == 5
    expected = ['This problem is.', 'It is therefore', 'However,we', 'without introducing ..', 'However, I have']
    assert [p for p in text.lines(t)] == expected


def test_foldcase():
    test = ' Fold THE cases of a text to lower casM'
    assert text.foldcase(test) == test.lower()


def test_nopunctuation():
    test = '''This problem is about sequence-bunching, %^$^%**^&*Â©Â©^(*&(*()()_+)_!@@#:><>>?/./,.,';][{}{]just'''
    expected = ['This', 'problem', 'is', 'about', 'sequence', 'bunching', 'Â', 'Â', 'just']
    assert text.nopunctuation(test).split() == expected

    test = 'This problem is about: sequence-bunching\n\n just \n'
    expected = 'This problem is about  sequence bunching   just  '
    assert text.nopunctuation(test) == expected


def test_unixlinesep():
    t = CR + LF + LF + CR + CR + LF
    assert text.unixlinesep(t) == LF + LF + LF + LF
    assert text.unixlinesep(t, True) == ' ' + LF + LF + LF + ' ' + LF


def test_nolinesep():
    t = CR + LF + CR + CR + CR + LF
    assert text.nolinesep(t) == '      '


def test_toascii():
    acc = u"ÀÁÂÃÄÅÇÈÉÊËÌÍÎÏÑÒÓÔÕÖØÙÚÛÜÝàáâãäåçèéêëìíîïñòóôõöøùúûüýÿẞß®©œŒØøÆæ₵₡￠¢Žž"
    expected = r'AAAAAACEEEEIIIINOOOOOUUUUYaaaaaaceeeeiiiinooooouuuuyyZz'
    assert text.toascii(acc, translit=False) == expected
    expected = r'AAAAAACEEEEIIIINOOOOOOUUUUYaaaaaaceeeeiiiinoooooouuuuyySsss(r)(c)oeOEOoAEae_CL/CC/Zz'
    assert text.toascii(acc, translit=True) == expected


def test_toascii_works_with_empty_unicode_or_bytes():
    assert text.toascii(b'', translit=False) == u''
    assert text.toascii(u'', translit=True) == u''
    assert text.toascii(b'', translit=False) == u''
    assert text.toascii(u'', translit=True) == u''


def test_python_safe_name():
    s = "not `\\a /`good` -safe name ??"
    assert text.python_safe_name(s) == 'not___a___good___safe_name'
    s1 = "string1++or+"
    s2 = "string1 +or "
    assert text.python_safe_name(s2) == text.python_safe_name(s1)


def test_as_unicode():
    assert text.as_unicode('') == ''
    assert isinstance(text.as_unicode(b'some bytes'), str)
    assert text.as_unicode(None) == None
    try:
        text.as_unicode(['foo'])
        raise Exception('Exception should have been raised')
    except AssertionError:
        pass



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os

from datetime import datetime

import commoncode.date

from commoncode import testcase


class TestDate(testcase.FileBasedTesting):

    def test_secs_from_epoch_can_handle_micro_and_nano_secs(self):
        test_file = self.get_temp_file()
        open(test_file, 'w').close()
        # setting modified time to desired values
        os.utime(test_file, (1301420665.046481, 1301420665.046481))
        # otherwise the issue does not happen (ie. on mac)
        if 1301420665.0 < os.stat(test_file).st_mtime:
            file_date = commoncode.date.get_file_mtime(test_file)
            commoncode.date.secs_from_epoch(file_date)

    def test_get_file_mtime_for_a_new_file(self):
        test_file = self.get_temp_file()
        open(test_file, 'w').close()

        def as_yyyymmdd(s):
            return s[:10]

        now = datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')
        result = commoncode.date.get_file_mtime(test_file)
        assert as_yyyymmdd(result) == as_yyyymmdd(now)

    def test_get_file_mtime_for_a_modified_file(self):
        test_file = self.get_temp_file()
        open(test_file, 'w').close()
        expected = u'1992-05-09 00:00:00'
        m_ts = (24 * 3600) * 134 + (24 * 3600 * 365) * 22
        # setting modified time to expected values
        os.utime(test_file, (m_ts, m_ts))
        assert commoncode.date.get_file_mtime(test_file) == expected

    def test_get_file_mtime_for_a_modified_file_2(self):
        test_file = self.get_temp_file()
        open(test_file, 'w').close()
        # setting modified time to expected values
        expected = u'2011-01-06 14:35:00'
        os.utime(test_file, (1294324500, 1294324500))
        assert commoncode.date.get_file_mtime(test_file) == expected



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import logging
import os

import requests
from requests.exceptions import ConnectionError
from requests.exceptions import InvalidSchema

from commoncode import fileutils

logger = logging.getLogger(__name__)
# import sys
# logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
# logger.setLevel(logging.DEBUG)


def download_url(url, file_name=None, verify=True, timeout=10):
    """
    Fetch `url` and return the temporary location where the fetched content was
    saved. Use `file_name` if provided or create a new `file_name` base on the last
    url segment. If `verify` is True, SSL certification is performed. Otherwise, no
    verification is done but a warning will be printed.
    `timeout` is the timeout in seconds.
    """
    requests_args = dict(timeout=timeout, verify=verify)
    file_name = file_name or fileutils.file_name(url)

    try:
        response = requests.get(url, **requests_args)
    except (ConnectionError, InvalidSchema) as e:
        logger.error('download_url: Download failed for %(url)r' % locals())
        raise

    status = response.status_code
    if status != 200:
        msg = 'download_url: Download failed for %(url)r with %(status)r' % locals()
        logger.error(msg)
        raise Exception(msg)

    tmp_dir = fileutils.get_temp_dir(prefix='fetch-')
    output_file = os.path.join(tmp_dir, file_name)
    with open(output_file, 'wb') as out:
        out.write(response.content)

    return output_file


def ping_url(url):
    """
    Returns True is `url` is reachable.
    """
    try:
        from urlib.request import urlopen
    except ImportError:
        from urllib2 import urlopen

    # FIXME: if there is no 200 HTTP status, then the ULR may not be reachable.
    try:
        urlopen(url)
    except Exception:
        return False
    else:
        return True



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

from itertools import chain

from commoncode import fileset
from commoncode import filetype
from commoncode import fileutils

"""
Support for ignoring some file patterns such as .git or .svn directories, used
typically when walking file systems.
Also handle .ignore-like file and provide common default ignores.
"""


def is_ignored(location, ignores, unignores=None, skip_special=True):
    """
    Return a tuple of (pattern , message) if a file at location is ignored
    or False otherwise.
    `ignores` and `unignores` are mappings of patterns to a reason.

    If `skip_special` is True, location is checked and ignored if is considered a special file,
    e.g. symlink, FIFO, device file, etc. and location is required to be an actual location to a
    file rather than a path string.
    """
    if skip_special and filetype.is_special(location):
        return True
    return not fileset.is_included(location, includes=unignores, excludes=ignores)


def is_ignore_file(location):
    """
    Return True if the location is an ignore file.
    """
    return (
        filetype.is_file(location)
        and fileutils.file_name(location) == '.scancodeignore'
    )


def get_ignores(location, include_defaults=True):
    """
    Return a ignores and unignores patterns mappings loaded from the
    file at `location`. Optionally include defaults patterns
    """
    ignores = {}
    unignores = {}
    if include_defaults:
        ignores.update(default_ignores)
    patterns = fileset.load(location)
    ign, uni = fileset.includes_excludes(patterns, location)
    ignores.update(ign)
    unignores.update(uni)
    return ignores, unignores

#
# Default ignores
#


ignores_MacOSX = {
    '.DS_Store': 'Default ignore: MacOSX artifact',
    '._.DS_Store': 'Default ignore: MacOSX artifact',
    '__MACOSX': 'Default ignore: MacOSX artifact',
    '.AppleDouble': 'Default ignore: MacOSX artifact',
    '.LSOverride': 'Default ignore: MacOSX artifact',
    '.DocumentRevisions-V100': 'Default ignore: MacOSX artifact',
    '.fseventsd': 'Default ignore: MacOSX artifact',
    '.Spotlight-V100': 'Default ignore: MacOSX artifact',
    '.VolumeIcon.icns': 'Default ignore: MacOSX artifact',

    '.journal': 'Default ignore: MacOSX DMG/HFS+ artifact',
    '.journal_info_block': 'Default ignore: MacOSX DMG/HFS+ artifact',
    '.Trashes': 'Default ignore: MacOSX DMG/HFS+ artifact',
    r'\[HFS+ Private Data\]': 'Default ignore: MacOSX DMG/HFS+ artifact private data',
}

ignores_Windows = {
    'Thumbs.db': 'Default ignore: Windows artifact',
    'ehthumbs.db': 'Default ignore: Windows artifact',
    'Desktop.ini': 'Default ignore: Windows artifact',
    '$RECYCLE.BIN': 'Default ignore: Windows artifact',
    '*.lnk': 'Default ignore: Windows artifact',
    'System Volume Information': 'Default ignore: Windows FS artifact',
    'NTUSER.DAT*': 'Default ignore: Windows FS artifact',
}

ignores_Linux = {
    '.directory': 'Default ignore: KDE artifact',
    '.Trash-*': 'Default ignore: Linux/Gome/KDE artifact',
}

ignores_IDEs = {
    '*.el': 'Default ignore: EMACS Elisp artifact',
    '*.swp': 'Default ignore: VIM artifact',
    '.project': 'Default ignore: Eclipse IDE artifact',
    '.pydevproject': 'Default ignore: Eclipse IDE artifact',
    '.settings': 'Default ignore: Eclipse IDE artifact',
    '.eclipse': 'Default ignore: Eclipse IDE artifact',
    '.loadpath': 'Default ignore: Eclipse IDE artifact',
    '*.launch': 'Default ignore: Eclipse IDE artifact',
    '.cproject': 'Default ignore: Eclipse IDE artifact',
    '.cdtproject': 'Default ignore: Eclipse IDE artifact',
    '.classpath': 'Default ignore: Eclipse IDE artifact',
    '.buildpath': 'Default ignore: Eclipse IDE artifact',
    '.texlipse': 'Default ignore: Eclipse IDE artifact',

    '*.iml': 'Default ignore: JetBrains IDE artifact',
    '*.ipr': 'Default ignore: JetBrains IDE artifact',
    '*.iws': 'Default ignore: JetBrains IDE artifact',
    '.idea/': 'Default ignore: JetBrains IDE artifact',
    '.idea_modules/': 'Default ignore: JetBrains IDE artifact',

    '*.kdev4': 'Default ignore: Kdevelop artifact',
    '.kdev4/': 'Default ignore: Kdevelop artifact',

    '*.nib': 'Default ignore: Apple Xcode artifact',
    '*.plst': 'Default ignore: Apple Xcode plist artifact',
    '*.pbxuser': 'Default ignore: Apple Xcode artifact',
    '*.pbxproj': 'Default ignore: Apple Xcode artifact',
    'xcuserdata': 'Default ignore: Apple Xcode artifact',
    '*.xcuserstate': 'Default ignore: Apple Xcode artifact',

    '*.csproj': 'Default ignore: Microsoft VS project artifact',
    '*.unityproj': 'Default ignore: Microsoft VS project artifact',
    '*.sln': 'Default ignore: Microsoft VS project artifact',
    '*.sluo': 'Default ignore: Microsoft VS project artifact',
    '*.suo': 'Default ignore: Microsoft VS project artifact',
    '*.user': 'Default ignore: Microsoft VS project artifact',
    '*.sln.docstates': 'Default ignore: Microsoft VS project artifact',
    '*.dsw': 'Default ignore: Microsoft VS project artifact',

    '.editorconfig': 'Default ignore: Editor config artifact',

    ' Leiningen.gitignore': 'Default ignore: Leiningen artifact',
    '.architect': 'Default ignore: ExtJS artifact',
    '*.tmproj': 'Default ignore: Textmate artifact',
    '*.tmproject': 'Default ignore: Textmate artifact',
}

ignores_web = {
    '.htaccess': 'Default ignore: .htaccess file',
    'robots.txt': 'Default ignore: robots file',
    'humans.txt': 'Default ignore: robots file',
    'web.config': 'Default ignore: web config',
    '.htaccess.sample': 'Default ignore: .htaccess file',
}

ignores_Maven = {
    'pom.xml.tag': 'Default ignore: Maven artifact',
    'pom.xml.releaseBackup': 'Default ignore: Maven artifact',
    'pom.xml.versionsBackup': 'Default ignore: Maven artifact',
    'pom.xml.next': 'Default ignore: Maven artifact',
    'release.properties': 'Default ignore: Maven artifact',
    'dependency-reduced-pom.xml': 'Default ignore: Maven artifact',
    'buildNumber.properties': 'Default ignore: Maven artifact',
}

ignores_VCS = {
    '.bzr': 'Default ignore: Bazaar artifact',
    '.bzrignore' : 'Default ignore: Bazaar config artifact',

    '.git': 'Default ignore: Git artifact',
    '.gitignore' : 'Default ignore: Git config artifact',
    '.gitattributes': 'Default ignore: Git config artifact',

    '.hg': 'Default ignore: Mercurial artifact',
    '.hgignore' : 'Default ignore: Mercurial config artifact',

    '.repo': 'Default ignore: Multiple Git repository artifact',

    '.svn': 'Default ignore: SVN artifact',
    '.svnignore': 'Default ignore: SVN config artifact',

    '.tfignore': 'Default ignore: Microsft TFS config artifact',

    'vssver.scc': 'Default ignore: Visual Source Safe artifact',

    'CVS': 'Default ignore: CVS artifact',
    '.cvsignore': 'Default ignore: CVS config artifact',

    '*/_MTN': 'Default ignore: Monotone artifact',
    '*/_darcs': 'Default ignore: Darcs artifact',
    '*/{arch}': 'Default ignore: GNU Arch artifact',
}

ignores_Medias = {
    'pspbrwse.jbf': 'Default ignore: Paintshop browse file',
    'Thumbs.db': 'Default ignore: Image thumbnails DB',
    'Thumbs.db:encryptable': 'Default ignore: Image thumbnails DB',
    'thumbs/': 'Default ignore: Image thumbnails DB',
    '_thumbs/': 'Default ignore: Image thumbnails DB',
}

ignores_Build_scripts = {
    'Makefile.in': 'Default ignore: automake artifact',
    'Makefile.am': 'Default ignore: automake artifact',
    'autom4te.cache': 'Default ignore: autoconf artifact',
    '*.m4': 'Default ignore: autotools artifact',
    'configure': 'Default ignore: Configure script',
    'configure.bat': 'Default ignore: Configure script',
    'configure.sh': 'Default ignore: Configure script',
    'configure.ac': 'Default ignore: Configure script',
    'config.guess': 'Default ignore: Configure script',
    'config.sub': 'Default ignore: Configure script',
    'compile': 'Default ignore: autoconf artifact',
    'depcomp': 'Default ignore: autoconf artifact',
    'ltmain.sh': 'Default ignore: libtool autoconf artifact',
    'install-sh': 'Default ignore: autoconf artifact',
    'missing': 'Default ignore: autoconf artifact',
    'mkinstalldirs': 'Default ignore: autoconf artifact',
    'stamp-h1': 'Default ignore: autoconf artifact',
    'm4/': 'Default ignore: autoconf artifact',
    'autogen.sh': 'Default ignore: autotools artifact',
    'autogen.sh': 'Default ignore: autotools artifact',

    'CMakeCache.txt': 'Default ignore: CMake artifact',
    'cmake_install.cmake': 'Default ignore: CMake artifact',
    'install_manifest.txt': 'Default ignore: CMake artifact',
}

ignores_CI = {
    '.travis.yml' : 'Default ignore: Travis config',
    '.coveragerc' : 'Default ignore: Coverall config',
}

ignores_Python = {
    'pip-selfcheck.json': 'Default ignore: Pip workfile',
    'pytest.ini': 'Default ignore: Python pytest config',
    'tox.ini': 'Default ignore: Python tox config',
    '__pycache__/': 'Default ignore: Python bytecode cache',
    '.installed.cfg': 'Default ignore: Python Buildout artifact',
    'pip-log.txt': 'Default ignore: Python pip artifact',
    'pip-delete-this-directory.txt': 'Default ignore: Python pip artifact',
    'pyvenv.cfg': 'Default ignore: Python virtualenv artifact',
}

ignores_I18N = {
    '*.mo': 'Default ignore: Translation file',
    '*.pot': 'Default ignore: Translation file',
    '.localized': 'Default ignore: localized file',
}

ignores_coverage_and_tests = {
    '*.gcno': 'Default ignore: GCC coverage',
    '*.gcda': 'Default ignore: GCC coverage',
    '*.gcov': 'Default ignore: GCC coverage',
    '.last_cover_stats': 'Default ignore: Perl coverage',
    'htmlcov/': 'Default ignore: Python coverage',
    '.tox/': 'Default ignore: Tox tem dir',
    '.coverage': 'Default ignore: Python coverage',
    '.coverage.*': 'Default ignore: Python coverage',
    'nosetests.xml': 'Default ignore: Python nose tests',
    'coverage.xml': 'Default ignore: Python coverage',
    '/spec/reports/': 'Default ignore: Ruby Rails test report',
    '/rdoc/': 'Default ignore: Ruby doc',
    '.rvmrc': 'Default ignore: Ruby RVM',
    '.sass-cache': 'Default ignore: Saas cache',
    '*.css.map': 'Default ignore: Saas map',
    'phpunit.xml': 'Default ignore: phpunit',
    '*.VisualState.xml': 'Default ignore: Nunit',
    'TestResult.xml': 'Default ignore: Nunit',
}

ignores_Misc = {
    'pax_global_header': 'Default ignore: Pax header file',
    'C++.gitignore': 'Default ignore: C++.gitignore',
    '.gwt/': 'Default ignore: GWT compilation logs',
    '.gwt-tmp/': 'Default ignore: GWT temp files',
    'gradle-app.setting': 'Default ignore: Graddle app settings',
    'hs_err_pid*': 'Default ignore: Java VM crash logs',
    '.grunt': 'Default ignore: Grunt intermediate storage',
    '.history': 'Default ignore: History file',
    '.~lock.*#': 'Default ignore: LibreOffice locks',
    '/.ssh': 'Default ignore: SSH configuration',
}

default_ignores = {}

default_ignores.update(chain(*[d.items() for d in [
    ignores_MacOSX,
    ignores_Windows,
    ignores_Linux,
    ignores_IDEs,
    ignores_web,
    ignores_Maven,
    ignores_VCS,
    ignores_Medias,
    ignores_Build_scripts,
    ignores_CI,
    ignores_Python,
    ignores_I18N,
    ignores_coverage_and_tests,
    ignores_Misc,
    ignores_Build_scripts,
]]))



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import os
from os.path import join
from os.path import exists
from unittest import skipIf

from commoncode import filetype
from commoncode import fileutils
from commoncode.system import on_posix
from commoncode.system import on_windows
from commoncode.system import py3

import commoncode.testcase
from commoncode.testcase import FileBasedTesting
from commoncode.testcase import make_non_readable
from commoncode.testcase import make_non_writable


class TypeTest(commoncode.testcase.FileBasedTesting):
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def test_get_size_on_file(self):
        test_file = self.get_test_loc('filetype/size/Image1.eps')
        assert filetype.get_size(test_file) == 12388

    def test_get_size_on_directory(self):
        test_dir = self.get_test_loc('filetype/size', copy=True)
        assert filetype.get_size(test_dir) == 12400

    def test_get_type(self):
        test_dir = self.extract_test_tar('filetype/types.tar', verbatim=True)
        results = []
        for root, dirs, files in os.walk(test_dir):
            for d in dirs:
                results.append((d, filetype.get_type(os.path.join(root, d))))
            for f in files:
                results.append((f, filetype.get_type(os.path.join(root, f))))

        expected = [
            ('5-DIRTYPE', 'd'),
            ('0-REGTYPE', 'f'),
            ('0-REGTYPE-TEXT', 'f'),
            ('0-REGTYPE-VEEEERY_LONG_NAME___________________________________'
             '______________________________________________________________'
             '____________________155', 'f'),
            ('1-LNKTYPE', 'f'),
            ('S-SPARSE', 'f'),
            ('S-SPARSE-WITH-NULLS', 'f')
        ]

        # symlinks and special files are not supported on win
        if on_posix:
            expected += [ ('2-SYMTYPE', 'l'), ('6-FIFOTYPE', 's'), ]

        try:
            assert sorted(results) == sorted(expected)
        except Exception as e:
            if on_windows and py3:
                # On some Windows symlinkes are detected OK (Windows 10?) but not in Windows 7
                expected += [ ('2-SYMTYPE', 'l') ]
                assert sorted(results) == sorted(expected)
            else:
                raise e

    def test_is_rwx_with_none(self):
        assert not filetype.is_writable(None)
        assert not filetype.is_readable(None)
        assert not filetype.is_executable(None)

    def test_is_readable_is_writeable_file(self):
        base_dir = self.get_test_loc('filetype/readwrite', copy=True)
        test_file = os.path.join(os.path.join(base_dir, 'sub'), 'file')

        try:
            assert filetype.is_readable(test_file)
            assert filetype.is_writable(test_file)

            make_non_readable(test_file)
            if on_posix:
                assert not filetype.is_readable(test_file)

            make_non_writable(test_file)
            assert not filetype.is_writable(test_file)
        finally:
            fileutils.chmod(base_dir, fileutils.RW, recurse=True)

    def test_is_readable_is_writeable_dir(self):
        base_dir = self.get_test_loc('filetype/readwrite', copy=True)
        test_dir = os.path.join(base_dir, 'sub')

        try:
            assert filetype.is_readable(test_dir)
            assert filetype.is_writable(test_dir)

            make_non_readable(test_dir)
            if on_posix:
                assert not filetype.is_readable(test_dir)
            else:
                # dirs are always RW on windows
                assert filetype.is_readable(test_dir)
            make_non_writable(test_dir)
            if on_posix:
                assert not filetype.is_writable(test_dir)
            else:
                # dirs are always RW on windows
                assert filetype.is_writable(test_dir)
            # finally
        finally:
            fileutils.chmod(base_dir, fileutils.RW, recurse=True)


class CountTest(FileBasedTesting):
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    def get_test_count_dir(self):
        test_dir = self.get_test_loc('count/filecount', copy=True)
        sub3 = join(test_dir, 'dir', 'sub3')
        if not exists(sub3):
            os.makedirs(sub3)
        return test_dir

    def test_get_file_count_with_empty_dir(self):
        test_dir = self.get_temp_dir()
        assert filetype.get_file_count(test_dir) == 0

    def test_get_file_count_with_single_file(self):
        test_file = self.get_temp_file()
        with open(test_file, 'w') as f:
            f.write(u'')
        assert filetype.is_file(test_file)
        assert filetype.get_file_count(test_file) == 1

    def test_get_file_count_with_empty_folders(self):
        test_dir = self.get_test_count_dir()
        result = filetype.get_file_count(test_dir)
        assert result == 9

    def test_get_file_size_and_count(self):
        test_dir = self.get_test_count_dir()
        result = filetype.get_size(test_dir)
        assert result == 18

    def test_get_file_size(self):
        test_dir = self.get_test_count_dir()
        tests = (
            ('dir/a.txt', 2),
            ('dir/b.txt', 2),
            ('dir/c.txt', 2),
            ('dir/sub1/a.txt', 2),
            ('dir/sub1/b.txt', 2),
            ('dir/sub1/c.txt', 2),
            ('dir/sub1/subsub/a.txt', 2),
            ('dir/sub1/subsub/b.txt', 2),
            ('dir/sub1/subsub', 4),
            ('dir/sub1', 10),
            ('dir/sub2/a.txt', 2),
            ('dir/sub2', 2),
            ('dir/sub3', 0),
            ('dir/', 18),
            ('', 18),
        )
        for test_file, size in tests:
            result = filetype.get_size(os.path.join(test_dir, test_file))
            assert result == size

    def test_get_file_count(self):
        test_dir = self.get_test_count_dir()
        tests = (
            ('dir/a.txt', 1),
            ('dir/b.txt', 1),
            ('dir/c.txt', 1),
            ('dir/sub1/a.txt', 1),
            ('dir/sub1/b.txt', 1),
            ('dir/sub1/c.txt', 1),
            ('dir/sub1/subsub/a.txt', 1),
            ('dir/sub1/subsub/b.txt', 1),
            ('dir/sub1/subsub', 2),
            ('dir/sub1', 5),
            ('dir/sub2/a.txt', 1),
            ('dir/sub2', 1),
            ('dir/sub3', 0),
            ('dir/', 9),
            ('', 9),
        )
        for test_file, count in tests:
            result = filetype.get_file_count(os.path.join(test_dir, test_file))
            assert result == count


class SymlinkTest(FileBasedTesting):
    test_data_dir = os.path.join(os.path.dirname(__file__), 'data')

    @skipIf(on_windows, 'os.symlink does not work on Windows')
    def test_is_file(self):
        test_file = self.get_test_loc('symlink/test', copy=True)
        temp_dir = fileutils.get_temp_dir()
        test_link = join(temp_dir, 'test-link')
        os.symlink(test_file, test_link)
        assert filetype.is_file(test_link, follow_symlinks=True)
        assert not filetype.is_file(test_link, follow_symlinks=False)

    @skipIf(on_windows, 'os.symlink does not work on Windows')
    def test_is_dir(self):
        test_dir = self.get_test_loc('symlink', copy=True)
        temp_dir = fileutils.get_temp_dir()
        test_link = join(temp_dir, 'test-dir-link')
        os.symlink(test_dir, test_link)
        assert filetype.is_dir(test_link, follow_symlinks=True)
        assert not filetype.is_dir(test_link, follow_symlinks=False)
